"""
php_ini - file ``/etc/php.ini``
===============================

This module provides the ``PHPConfig`` class parser, for reading the
options in the ``/etc/php.ini`` file.

Typical content of ``/etc/php.ini`` file is::

    [PHP]
    engine = On
    short_open_tag = Off
    precision = 14
    output_buffering = 4096
    zlib.output_compression = Off
    implicit_flush = Off
    unserialize_callback_func =
    serialize_precision = -1
    disable_functions =
    disable_classes =
    zend.enable_gc = On
    zend.exception_ignore_args = On
    zend.exception_string_param_max_len = 0
    expose_php = On
    max_execution_time = 30
    max_input_time = 60
    memory_limit = 128M
    error_reporting = E_ALL & ~E_DEPRECATED & ~E_STRICT
    default_mimetype = "text/html"

The class has one attribute ``data`` which is a nested ``dict`` representing sections
of the input INI file. Each section is represented as ``dict`` where keys are name
of options and values are values of those options.

Example:
    >>> php_conf["PHP"]["default_mimetype"].value
    'text/html'
    >>> php_config.data['PHP']['default_mimetype']
    'text/html'
    >>> php_conf.data['Session']['session.cache_limiter']
    'nocache'
    >>> php_conf["PHP"]["max_execution_time"].value
     30
    >>> php_conf["PHP"]["engine"].value  # 'On' turns to 'True'
    True
    >>> php_conf["PHP"]["short_opeh_tag"].value  # 'Off' turns to 'False'
    False
    >>> php_c['PHP']['precision'].value
    14
    >>> php_conf.get("PHP").get("memory_limit")  # '128M' is converted into bytes
    134217728
"""
import string

from insights.core import ConfigParser
from insights.core.filters import add_filter
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsr import (Char, EOF, HangingString, InSet, LeftBracket,
                            Lift, LineEnd, Literal, Many, Number,
                            OneLineComment, Opt, PosMarker, QuotedString,
                            RightBracket, String, WithIndent, WS, WSChar,
                            skip_none)
from insights.parsr.query import Directive, Entry, Section, eq
from insights.specs import Specs


# Filter to ensure that the section headings will always be included.
add_filter(Specs.php_ini, "[")


@parser(Specs.php_ini, continue_on_error=False)
class PHPConf(ConfigParser):
    """
    Class for php configuration file.
    """
    def parse_doc(self, content):
        try:
            def to_directive(x):
                name, rest = x
                rest = [rest] if rest is not None else []
                return Directive(name=name.value.strip(), attrs=rest, lineno=name.lineno, src=self)

            def to_section(name, rest):
                return Section(name=name.value.strip(), children=rest, lineno=name.lineno, src=self)

            def apply_defaults(cfg):
                if "DEFAULT" not in cfg:
                    return cfg

                defaults = cfg["DEFAULT"]
                not_defaults = cfg[~eq("DEFAULT")]
                for c in not_defaults:
                    for d in defaults.grandchildren:
                        if d.name not in c:
                            c.children.append(d)

                cfg.children = list(not_defaults)
                return cfg

            def make_bytes(number, char_multiple):
                if char_multiple.lower() == 'k':
                    return number * 2**10
                if char_multiple.lower() == 'm':
                    return number * 2**20
                if char_multiple.lower() == 'g':
                    return number * 2**30

            content = "\n".join(content)

            header_chars = (set(string.printable) - set(string.whitespace) - set("[]")) | set(" ")
            sep_chars = set("=")
            key_chars = header_chars - sep_chars
            value_chars = set(string.printable) - set("\n\r")

            On = Literal("on", True, ignore_case=True)
            Off = Literal("off", False, ignore_case=True)
            Tru = Literal("true", True, ignore_case=True)
            Fals = Literal("false", False, ignore_case=True)
            Boolean = ((On | Off | Tru | Fals) & (WSChar | LineEnd)) % "Boolean"
            Num = Number & (WSChar | LineEnd)
            QuoStr = QuotedString & (WSChar | LineEnd)
            # Handle php.ini shorthand notation for memory limits: 1G, 8M, 50K
            # https://www.php.net/manual/en/faq.using.php#faq.using.shorthandbytes
            MemNum = (Lift(make_bytes) * Number * (Char('K') | Char('M') | Char('G'))) & (WSChar | LineEnd)

            LeftEnd = (WS + LeftBracket + WS)
            RightEnd = (WS + RightBracket + WS)
            Header = (LeftEnd >> PosMarker(String(header_chars)) << RightEnd) % "Header"
            Key = WS >> PosMarker(String(key_chars)) << WS
            Sep = InSet(sep_chars, "Sep")
            Value = WS >> (Boolean | MemNum | Num | QuoStr | HangingString(value_chars))
            KVPair = WithIndent(Key + Opt(Sep >> Value)) % "KVPair"
            Comment = (WS >> (OneLineComment(";")).map(lambda x: None))

            Line = Comment | KVPair.map(to_directive)
            Sect = Lift(to_section) * Header * Many(Line).map(skip_none)
            Doc = Many(Comment | Sect).map(skip_none)
            Top = Doc << WS << EOF

            res = Entry(children=Top(content), src=self)
            return apply_defaults(res)

        except SkipComponent:
            raise
        except:
            raise ParseException("Could not parse content: '{0}'".
                                                format(content))

    def parse_content(self, content):
        super(PHPConf, self).parse_content(content)
        dict_all = {}
        for section in self.doc:
            section_dict = {}
            option_names = set(o.name for o in section)
            for name in option_names:
                options = [str(o.value) for o in section[name]]
                section_dict[name] = options[0] if len(options) == 1 else options
            dict_all[section.name] = section_dict
        self.data = dict_all



"""
VirtUuidFacts - files ``/etc/rhsm/facts/virt_uuid.facts``
=========================================================

This module provides parsing for the ``/etc/rhsm/facts/virt_uuid.facts`` file.
The ``VirtUuidFacts`` class is based on a shared class which processes the JSON
information into a dictionary.

Sample input data looks like::

    {"virt.uuid": "4546B285-6C41-5D6R-86G5-0BFR4B3625FS", "uname.machine": "x86"}

Examples:

    >>> len(virt_uuid_facts.data)
    2
"""

from insights.specs import Specs

from .. import JSONParser, parser


@parser(Specs.virt_uuid_facts)
class VirtUuidFacts(JSONParser):
    pass



"""
QemuConf - file ``/etc/libvirt/qemu.conf``
==========================================

The ``/etc/libvirt/qemu.conf`` file is in a key-value format, but there are several lines for one value.

Given a file containing the following test data::

   vnc_listen = "0.0.0.0"
   vnc_auto_unix_socket = 1
   vnc_tls = 1
   vnc_tls_x509_cert_dir = "/etc/pki/libvirt-vnc"
   security_driver = "selinux"
   cgroup_device_acl = [
    "/dev/null", "/dev/full", "/dev/zero",
    "/dev/random", "/dev/urandom",
    "/dev/ptmx", "/dev/kvm", "/dev/kqemu",
    "/dev/rtc","/dev/hpet", "/dev/vfio/vfio"
    ]

Example:
    >>> config = shared[QemuConf]
    >>> config.get('vnc_listen')
    '0.0.0.0'
    >>> config.get('vnc_tls')
    '1'
    >>> "/dev/random" in config.get('cgroup_device_acl')
    True
"""
import json
from .. import parser, Parser, get_active_lines, LegacyItemAccess
from insights.specs import Specs


@parser(Specs.qemu_conf)
class QemuConf(LegacyItemAccess, Parser):
    """
    A dict of the content of the ``qemu.conf`` configuration file.

    Attributes:
        data (dict): Dictionary of parsed data that splits by '='
    """

    def parse_content(self, content):
        """Parse file content of qemu.conf"""
        _lines = get_active_lines(content, comment_char="#")
        self.data = {}
        current_key = ""
        for line in _lines:
            if "=" in line:
                k, v = line.split('=')
                current_key = k.strip()
                self.data[current_key] = v.strip("\"' ")
            else:
                self.data[current_key] = self.data[current_key] + line
            # if the value like "[ , ,]", convert to list
            if self.data[current_key].endswith("]"):
                self.data[current_key] = json.loads(self.data[current_key])



from insights.parsers import buddyinfo
from insights.parsers.buddyinfo import BuddyInfo
from insights.core.exceptions import SkipComponent
from insights.tests import context_wrap

import doctest
import pytest

BUDDYINFO_1 = """
Node 0, zone      DMA      0      0      0      0      0      0      0      0      1      1      2
Node 0, zone    DMA32      8     10     12      9      9     11     10     12     12     11    569
Node 0, zone   Normal    460    485    375   1611   3201   2187   1437    844    487    247   6120
Node 1, zone   Normal      1      7   1783   2063   1773   3227   2299   1399    729    430   6723
""".strip()

BUDDYINFO_2 = """
Node 0, zone      DMA      0      0      0      0      1      0      0      1      0      1      2
Node 0, zone    DMA32    890    610    792    663    464    285    128     96    155      9      3
Node 0, zone   Normal  19340   9441   1865    250    163     47      3      3      1      3      0
""".strip()


def test_buddyinfo():
    buddy = BuddyInfo(context_wrap(BUDDYINFO_1))
    assert len(buddy) == 4

    mem = buddy[2]
    assert mem['node'] == '0'
    assert mem['zone'] == 'Normal'
    assert len(mem['counter']) == 11
    assert mem['counter'][10] == 6120
    assert "Node 0, zone   Normal    460    485" in mem['raw']

    mem = buddy[3]
    assert mem['node'] == '1'
    assert mem['zone'] == 'Normal'
    assert mem['counter'][10] == 6723
    assert "Node 1, zone   Normal      1      7" in mem['raw']

    buddy = BuddyInfo(context_wrap(BUDDYINFO_2))
    assert len(buddy) == 3
    assert buddy[0]['counter'][9] == 1

    with pytest.raises(SkipComponent):
        BuddyInfo(context_wrap(""))


def test_doc_examples():
    env = {
        'buddy': BuddyInfo(context_wrap(BUDDYINFO_1))
    }
    failed, total = doctest.testmod(buddyinfo, globs=env)
    assert failed == 0



"""
GCPInstanceType
===============

This parser simply reads the output of command
``curl http://metadata.google.internal/computeMetadata/v1/instance/machine-type -H 'Metadata-Flavor: Google'``,
which is used to check the machine type of the Google instance of the host.

For more details, See:
- https://cloud.google.com/compute/docs/machine-types
- https://cloud.google.com/compute/docs/storing-retrieving-metadata#api_4

"""
from insights.core import CommandParser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.gcp_instance_type)
class GCPInstanceType(CommandParser):
    """
    Class for parsing the GCP Instance type returned by command
    ``curl -s -H 'Metadata-Flavor: Google' http://metadata.google.internal/computeMetadata/v1/instance/machine-type``,


    Typical output of this command is::

        projects/123456789/machineTypes/n2-highcpu-16


    Raises:
        SkipComponent: When content is empty or no parse-able content.
        ParseException: When type cannot be recognized.

    Attributes:
        type (str): The type of VM instance in GCP, e.g: n2
        size (str): The size of VM instance in GCP, e.g: highcpu-16
        raw (str): The fully type string, e.g. 'n2-highcpu-16'
        raw_line (str): The fully type string returned by the ``curl`` command

    Examples:
        >>> gcp_inst.type
        'n2'
        >>> gcp_inst.size
        'highcpu-16'
        >>> gcp_inst.raw
        'n2-highcpu-16'
    """

    def parse_content(self, content):
        if not content or 'curl: ' in content[0]:
            raise SkipComponent()

        self.raw_line = self.raw = self.type = self.size = None
        # Ignore any curl stats that may be present in data
        for l in content:
            l_strip = l.strip()
            if ' ' not in l_strip and '-' in l_strip:
                self.raw_line = l_strip
                self.raw = l_strip.split('/')[-1]
                type_sp = self.raw.split('-', 1)
                self.type, self.size = type_sp[0], type_sp[1]

        if not self.type:
            raise ParseException('Unrecognized type: "{0}"', content[0])

    def __repr__(self):
        return "<gcp_type: {t}, size: {s}, raw: {r}>".format(
                t=self.type, s=self.size, r=self.raw_line)



from insights.parsers import nfs_exports as parsermodule
from insights.parsers.nfs_exports import NFSExports, NFSExportsD
from insights.tests import context_wrap
from doctest import testmod

EXPORTS = """
/home/utcs/shared/ro                    @group(ro,sync)   ins1.example.com(rw,sync,no_root_squash) ins2.example.com(rw,sync,no_root_squash)
/home/insights/shared/rw                @group(rw,sync)   ins1.example.com(rw,sync,no_root_squash) ins2.example.com(ro,sync,no_root_squash)
/home/insights/shared/special/all/mail  @group(rw,sync,no_root_squash)
/home/insights/ins/special/all/config   @group(ro,sync,no_root_squash)  ins1.example.com(rw,sync,no_root_squash)
#/home/insights                          ins1.example.com(rw,sync,no_root_squash)
/home/example                           @group(rw,sync,root_squash) ins1.example.com(rw,sync,no_root_squash) ins2.example.com(rw,sync,no_root_squash)
# A duplicate host for this exported path
/home/example                           ins2.example.com(rw,sync,no_root_squash)
"""


def test_module_documentation():
    failed, total = testmod(parsermodule, globs={
        "exports": NFSExports(context_wrap(EXPORTS))
    })
    assert failed == 0


def test_nfs_exports():
    nfs_exports = NFSExports(context_wrap(EXPORTS))
    _do(nfs_exports)


def test_nfs_exports_d():
    nfs_exports = NFSExportsD(context_wrap(EXPORTS))
    _do(nfs_exports)


def test_nfs_exports_empty():
    nfs_exports = NFSExports(context_wrap(""))
    _do_empty(nfs_exports)


def test_nfs_exports_d_empty():
    nfs_exports = NFSExportsD(context_wrap(""))
    _do_empty(nfs_exports)


def _do_empty(nfs_exports):
    assert nfs_exports.data == {}
    assert nfs_exports.ignored_exports == {}
    assert nfs_exports.raw_lines == {}
    assert nfs_exports.all_options() == set()
    assert nfs_exports.export_paths() == set()


def _do(nfs_exports):
    assert nfs_exports.data == {
        "/home/utcs/shared/ro": {
            "@group": ["ro", "sync"],
            "ins1.example.com": ["rw", "sync", "no_root_squash"],
            "ins2.example.com": ["rw", "sync", "no_root_squash"]
        }, "/home/insights/shared/rw": {
            "@group": ["rw", "sync"],
            "ins1.example.com": ["rw", "sync", "no_root_squash"],
            "ins2.example.com": ["ro", "sync", "no_root_squash"]
        }, "/home/insights/shared/special/all/mail": {
            "@group": ["rw", "sync", "no_root_squash"]
        }, "/home/insights/ins/special/all/config": {
            "@group": ["ro", "sync", "no_root_squash"],
            "ins1.example.com": ["rw", "sync", "no_root_squash"]
        }, "/home/example": {
            "@group": ["rw", "sync", "root_squash"],
            "ins1.example.com": ["rw", "sync", "no_root_squash"],
            "ins2.example.com": ["rw", "sync", "no_root_squash"]
        }
    }

    assert nfs_exports.ignored_exports == {
        '/home/example': {'ins2.example.com': ['rw', 'sync', 'no_root_squash']}
    }

    assert nfs_exports.raw_lines == {
        "/home/utcs/shared/ro": [
            '/home/utcs/shared/ro                    @group(ro,sync)   ins1.example.com(rw,sync,no_root_squash) ins2.example.com(rw,sync,no_root_squash)'
        ], "/home/insights/shared/rw": [
            '/home/insights/shared/rw                @group(rw,sync)   ins1.example.com(rw,sync,no_root_squash) ins2.example.com(ro,sync,no_root_squash)'
        ], "/home/insights/shared/special/all/mail": [
            '/home/insights/shared/special/all/mail  @group(rw,sync,no_root_squash)'
        ], "/home/insights/ins/special/all/config": [
            '/home/insights/ins/special/all/config   @group(ro,sync,no_root_squash)  ins1.example.com(rw,sync,no_root_squash)'
        ], "/home/example": [
            '/home/example                           @group(rw,sync,root_squash) ins1.example.com(rw,sync,no_root_squash) ins2.example.com(rw,sync,no_root_squash)',
            '/home/example                           ins2.example.com(rw,sync,no_root_squash)'
        ]
    }

    assert nfs_exports.all_options() == set(["ro", "rw", "sync", "no_root_squash", "root_squash"])
    assert nfs_exports.export_paths() == set([
        "/home/utcs/shared/ro", "/home/insights/shared/rw",
        "/home/insights/shared/special/all/mail",
        "/home/insights/ins/special/all/config", "/home/example"
    ])


NFS_EXPORTS_CORNER_CASES = '''
# Host with default share options
/mnt/share      host1
# Export with two non-overlapping host definitions
/mnt/share      host2(rw)
# No matter how many times a host is listed, only the first gets picked up
/mnt/share      host2(ro)
# Need to list on separate lines because of dictionary key merging
/mnt/share      host2(no_root_squash)
'''


def test_nfs_exports_corner_cases():
    exports = NFSExports(context_wrap(NFS_EXPORTS_CORNER_CASES))
    assert exports

    assert exports.data['/mnt/share'] == {'host1': [], 'host2': ['rw']}
    # Note: only the last ignored export gets stored per host.
    assert exports.ignored_exports['/mnt/share'] == {'host2': ['no_root_squash']}
    for path, hosts in exports:
        assert path in exports.data
        assert exports.data[path] == hosts



"""
The plugins module defines the components used by the rest of insights and
specializes their interfaces and execution model where required.

This module includes the following :class:`CompoentType` subclasses:

    - :class:`datasource`
    - :class:`parser`
    - :class:`combiner`
    - :class:`rule`
    - :class:`condition`
    - :class:`incident`
    - :class:`fact`

It also contains the following :class:`Response` subclasses that :class:`rules`
may return:

    - :class:`make_pass`
    - :class:`make_response` (alias for make_fail)
    - :class:`make_fail`
    - :class:`make_info`
    - :class:`make_metadata`
    - :class:`make_metadata_key`
    - :class:`make_fingerprint`

"""
from __future__ import print_function

import logging
import signal
import traceback

from pprint import pformat
from six import StringIO

from insights import settings
from insights.core import dr
from insights.core.context import HostContext
from insights.core.exceptions import (CalledProcessError, ContentException, SkipComponent, TimeoutException,
                                      ValidationException)

log = logging.getLogger(__name__)


class PluginType(dr.ComponentType):
    """
    PluginType is the base class of plugin types like datasource, rule, etc.
    It provides a default invoke method that catches exceptions we don't
    want bubbling to the top of the evaluation loop. These exceptions are
    commonly raised by datasource components but could be in the context of any
    component since most datasource runtime errors are lazy.

    It's possible for a datasource to "succeed" and return an object but for an
    exception to be raised when the parser tries to access the content of that
    object. For example, when a command datasource is evaluated, it only checks
    that the command exists and is executable. Invocation of the command itself
    is delayed until the parser asks for its value. This helps with performance
    and memory consumption.
    """
    def invoke(self, broker):
        try:
            return super(PluginType, self).invoke(broker)
        except ContentException as ce:
            log.debug(ce)
            broker.add_exception(self.component, ce, traceback.format_exc())
            raise SkipComponent()
        except CalledProcessError as cpe:
            log.debug(cpe)
            broker.add_exception(self.component, cpe, traceback.format_exc())
            raise SkipComponent()


class component(PluginType):
    pass


class datasource(PluginType):
    """
    Decorates a component that one or more :class:`insights.core.Parser`
    subclasses will consume.
    """
    filterable = False
    multi_output = False
    raw = False

    def _handle_timeout(self, signum, frame):
        raise TimeoutException("Datasource spec {ds_name} timed out after {secs} seconds!".format(
            ds_name=dr.get_name(self.component), secs=self.timeout))

    def invoke(self, broker):
        # Grab the timeout from the decorator, or use the default of 120.

        if HostContext in broker:
            self.timeout = getattr(self, "timeout", 120)
            signal.signal(signal.SIGALRM, self._handle_timeout)
            signal.alarm(self.timeout)
        try:
            return self.component(broker)
        except ContentException as ce:
            log.debug(ce)
            ce_tb = traceback.format_exc()
            for reg_spec in dr.get_registry_points(self.component):
                broker.add_exception(reg_spec, ce, ce_tb)
            raise SkipComponent()
        except CalledProcessError as cpe:
            log.debug(cpe)
            cpe_tb = traceback.format_exc()
            for reg_spec in dr.get_registry_points(self.component):
                broker.add_exception(reg_spec, cpe, cpe_tb)
            raise SkipComponent()
        except TimeoutException as te:
            log.debug(te)
            te_tb = traceback.format_exc()
            for reg_spec in dr.get_registry_points(self.component):
                broker.add_exception(reg_spec, te, te_tb)
            raise SkipComponent()
        finally:
            if HostContext in broker:
                signal.alarm(0)


class parser(PluginType):
    """
    Decorates a component responsible for parsing the output of a
    :class:`datasource`. ``@parser`` should accept multiple arguments, the first
    will ALWAYS be the datasource the parser component should handle.
    Any subsequent argument will be a ``component`` used to determine if
    the parser should fire.
    ``@parser`` should only decorate subclasses of :class:`insights.core.Parser`.

    .. warning::
        If a Parser component handles a datasource that returns a ``list``, a
        Parser instance will be created for each element of the list. Combiners
        or rules that depend on the Parser will be passed the list of instances
        and **not** a single parser instance. By default, if any parser in the
        list succeeds, those parsers are passed on to dependents, even if
        others fail. If all parsers should succeed or fail together, pass
        ``continue_on_error=False``.
    """
    def __init__(self, *args, **kwargs):
        group = kwargs.get('group', dr.GROUPS.single)
        self.continue_on_error = kwargs.get('continue_on_error', True)
        super(parser, self).__init__(*args, group=group)

    def invoke(self, broker):
        dep_value = broker[self.requires[0]]
        exception = False

        if not isinstance(dep_value, list):
            try:
                return self.component(dep_value)
            except ContentException as ce:
                log.debug(ce)
                broker.add_exception(self.component, ce, traceback.format_exc())
                exception = True
            except CalledProcessError as cpe:
                log.debug(cpe)
                broker.add_exception(self.component, cpe, traceback.format_exc())
                exception = True

        if exception:
            raise SkipComponent()

        results = []
        for d in dep_value:
            try:
                r = self.component(d)
                if r is not None:
                    results.append(r)
            except ContentException as ce:
                log.debug(ce)
                broker.add_exception(self.component, ce, traceback.format_exc())
                if not self.continue_on_error:
                    exception = True
                    break
            except SkipComponent as sc:
                if broker.store_skips:
                    log.warning(sc)
                    broker.add_exception(component, sc, traceback.format_exc())
                else:
                    pass
            except CalledProcessError as cpe:
                log.debug(cpe)
                broker.add_exception(self.component, cpe, traceback.format_exc())
                if not self.continue_on_error:
                    exception = True
                    break
            except Exception as ex:
                tb = traceback.format_exc()
                log.warning(tb)
                broker.add_exception(self.component, ex, tb)
                if not self.continue_on_error:
                    exception = True
                    break

        if exception:
            raise SkipComponent()

        if not results:
            log.debug("All failed: %s" % dr.get_name(self.component))
            raise SkipComponent()

        return results


class metadata(parser):
    """
    Used for old cluster uber-archives.

    .. deprecated:: 1.x

    .. warning::
        Do not use this component type.
    """
    requires = ["metadata.json"]


class combiner(PluginType):
    """
    A decorator for a component that composes or "combines" other components.

    A typical use case is hiding slight variations in related parser
    interfaces. Another use case is to combine several related parsers behind a
    single, cohesive, higher level interface.
    """
    pass


class remoteresource(PluginType):
    """ ComponentType for a component for remote web resources. """
    pass


class rule(PluginType):
    """
    Decorator for components that encapsulate some logic that depends on the
    data model of a system. Rules can depend on :class:`datasource` instances,
    :class:`parser` instances, :class:`combiner` instances, or anything else.

    For example:

    .. code-block:: python

       @rule(SshDConfig, InstalledRpms, [ChkConfig, UnitFiles], optional=[IPTables, IpAddr])
       def report(sshd_config, installed_rpms, chk_config, unit_files, ip_tables, ip_addr):
           # ...
           # ... some complicated logic
           # ...
           bash = installed_rpms.newest("bash")
           return make_pass("BASH", bash=bash)

    Notice that the arguments to ``report`` correspond to the dependencies in
    the ``@rule`` decorator and are in the same order.

    Parameters to the decorator have these forms:

    ============  ===============================  ==========================
    Criteria      Example Decorator Arguments      Description
    ============  ===============================  ==========================
    Required      ``SshDConfig, InstalledRpms``    Regular arguments
    At Least One  ``[ChkConfig, UnitFiles]``       An argument as a list
    Optional      ``optional=[IPTables, IpAddr]``  A list following optional=
    ============  ===============================  ==========================

    If a parameter is required, the value provided for it is guaranteed not to
    be ``None``. In the example above, ``sshd_config`` and ``installed_rpms``
    will not be ``None``.

    At least one of the arguments to parameters of an "at least one"
    list will not be ``None``. In the example, either or both of ``chk_config``
    and ``unit_files`` will not be ``None``.

    Any or all arguments for optional parameters may be ``None``.

    The following keyword arguments may be passed to the decorator:

    Keyword Args:
        requires (list) **deprecated**: a list of components that all
            components decorated with this type will require. Instead of using
            ``requires=[...]``, just pass dependencies as variable arguments
            to ``@rule`` as in the example above.
        optional (list): a list of components that all components decorated with
            this type will implicitly depend on optionally. Additional components
            passed as ``optional`` to the decorator will be appended to this list.
        metadata (dict): an arbitrary dictionary of information to associate
            with the component you're decorating. It can be retrieved with
            ``get_metadata``.
        tags (list): a list of strings that categorize the component. Useful for
            formatting output or sifting through results for components you care
            about.
        group: ``GROUPS.single`` or ``GROUPS.cluster``. Used to organize
            components into "groups" that run together with :func:`insights.core.dr.run`.
        cluster (bool): if ``True`` will put the component into the
            ``GROUPS.cluster`` group. Defaults to ``False``. Overrides ``group``
            if ``True``.
        content (string or dict): a jinja2 template or dictionary of jinja2
            templates. The :class:`Response` subclasses rules can return are
            dictionaries. :class:`make_pass`, :class:`make_fail`, and
            :class:`make_response` all accept first a key and then a list of
            arbitrary keyword arguments. If content is a dictionary, the key is
            used to look up the template that the rest of the keyword argments
            will be interpolated into. If content is a string, then it is used
            for all return values of the rule. If content isn't defined but a
            ``CONTENT`` variable is declared in the module, it will be used for
            every rule in the module and also can be a string or list of
            dictionaries
        links (dict): a dictionary with strings as keys and lists of urls as
            values. The keys categorize the urls, e.g. "kcs" for kcs urls and
            "bugzilla" for bugzilla urls.
    """
    content = None
    links = None

    def __init__(self, *args, **kwargs):
        super(rule, self).__init__(*args, **kwargs)
        self.content = kwargs.get("content")
        self.links = kwargs.get("links")

    def process(self, broker):
        """
        Ensures dependencies have been met before delegating to `self.invoke`.
        """
        if any(i in broker for i in dr.IGNORE.get(self.component, [])):
            raise SkipComponent()
        missing = self.get_missing_dependencies(broker)
        if missing:
            return _make_skip(dr.get_name(self.component), missing)
        r = self.invoke(broker)
        if r is None:
            return make_none()
        if not isinstance(r, Response):
            raise Exception("rules must return Response objects.")
        return r


class condition(PluginType):
    """
    ComponentType used to encapsulate boolean logic you'd like to have analyzed
    by a rule analysis system. Conditions should return truthy values. ``None``
    is also a valid return type for conditions, so ``rules`` that depend on
    ``conditions`` that might return None should check their validity.
    """
    pass


class incident(PluginType):
    """
    ComponentType for a component used by rules that allows automated
    statistical analysis.
    """
    pass


class fact(PluginType):
    """
    ComponentType for a component that surfaces a dictionary or list of
    dictionaries that will be used later by cluster rules. The data from a fact
    is converted to a pandas Dataframe
    """
    pass


def is_type(component, _type):
    try:
        return issubclass(dr.get_component_type(component), _type)
    except:
        return False


def is_datasource(component):
    return is_type(component, datasource)


def is_parser(component):
    return is_type(component, parser)


def is_combiner(component):
    return is_type(component, combiner)


def is_rule(component):
    return is_type(component, rule)


def is_component(obj):
    return bool(dr.get_component_type(obj))


class Response(dict):
    """
    Response is the base class of response types that can be returned from
    rules.

    Subclasses must call __init__ of this class via super() and must provide
    the response_type class attribute.

    The key_name class attribute is optional, but if one is specified, the
    first argument to __init__ must not be None. If key_name is None, then
    the first argument to __init__ should be None. It's best to override
    __init__ in subclasses so users aren't required to pass None explicitly.
    """

    response_type = None
    """
    response_type is something like 'rule', 'metadata', 'fingerprint', etc. It
    is how downstream systems identify the type of information returned by a
    rule.
    """

    key_name = None
    """
    key_name is something like 'error_key', 'fingerprint_key', etc.  It is the
    key downstream systems use to look up the exact response returned by a
    rule.
    """

    def __init__(self, key, **kwargs):
        self.validate_kwargs(kwargs)

        r = {"type": self.response_type}
        if self.key_name:
            self.validate_key(key)
            r[self.key_name] = key

        kwargs.update(r)
        kwargs = self.adjust_for_length(key, r, kwargs)
        super(Response, self).__init__(kwargs)

    def get_key(self):
        """
        Helper function that uses the response's key_name to look up the
        response identifier. For a rule, this is like
        response.get("error_key").
        """
        if self.key_name:
            return self.get(self.key_name)

    def validate_kwargs(self, kwargs):
        """
        Validates expected subclass attributes and constructor keyword
        arguments.
        """
        if not self.response_type:
            msg = "response_type must be set on the Response subclass."
            raise ValidationException(msg)

        if (self.key_name and self.key_name in kwargs) or "type" in kwargs:
            name = self.__class__.__name__
            msg = "%s is an invalid argument for %s" % (self.key_name, name)
            raise ValidationException(msg)

    def validate_key(self, key):
        """ Called if the key_name class attribute is not None. """
        if not key:
            name = self.__class__.__name__
            msg = "%s response missing %s" % (name, self.key_name)
            raise ValidationException(msg, self)
        elif not isinstance(key, str):
            msg = "Response contains invalid %s type" % self.key_name
            raise ValidationException(msg, type(key))

    def adjust_for_length(self, key, r, kwargs):
        """
        Converts the response to a string and compares its length to a max
        length specified in settings. If the response is too long, an error is
        logged, and an abbreviated response is returned instead.
        """
        length = len(str(kwargs))
        if length > settings.defaults["max_detail_length"]:
            self._log_length_error(key, length)
            r["max_detail_length_error"] = length
            return r
        return kwargs

    def _log_length_error(self, key, length):
        """ Helper function for logging a response length error. """
        extra = {
            "max_detail_length": settings.defaults["max_detail_length"],
            "len": length
        }
        if self.key_name:
            extra[self.key_name] = key
        msg = "Length of data in %s is too long." % self.__class__.__name__
        log.error(msg, extra=extra)

    def __str__(self):
        key_val = self.get_key()
        keys = sorted(self)
        if self.key_name in keys:
            keys.remove(self.key_name)
        if "type" in keys:
            keys.remove("type")

        buf = StringIO()
        if not keys:
            print(key_val, file=buf)
            buf.seek(0)
            return buf.read()

        print("%s:" % key_val, file=buf)
        indent = len(max(keys, key=len)) + 6
        hang_indent = "\n" + " " * indent
        for k in keys:
            key = ("    %s" % k) + " " * (indent - len(k) - 6) + ": "
            buf.write(key)
            lines = pformat(self[k]).splitlines()
            num_lines = len(lines)
            if num_lines > 10:
                lines = lines[:10]
                lines.append("<...%s more lines...>" % (num_lines - 10))
            out = hang_indent.join(lines)
            buf.write(out)
            buf.write("\n")

        buf.seek(0)
        return buf.read()


class make_response(Response):
    """
    Returned by a rule to signal that its conditions have been met.

    Example:

    .. code-block:: python

        # completely made up package
        buggy = InstalledRpms.from_package("bash-3.4.23-1.el7")

        @rule(InstalledRpms)
        def report(installed_rpms):
           bash = installed_rpms.newest("bash")
           if bash == buggy:
               return make_response("BASH_BUG_123", bash=bash)
           return make_pass("BASH", bash=bash)

    .. deprecated:: 1.x

        Use :class:`make_fail` instead.
    """

    response_type = "rule"
    key_name = "error_key"


class make_fail(make_response):
    """
    Returned by a rule to signal that its conditions have been met.

    Example:

    .. code-block:: python

        # completely made up package
        buggy = InstalledRpms.from_package("bash-3.4.23-1.el7")

        @rule(InstalledRpms)
        def report(installed_rpms):
           bash = installed_rpms.newest("bash")
           if bash == buggy:
               return make_fail("BASH_BUG_123", bash=bash)
           return make_pass("BASH", bash=bash)
    """
    pass


class make_pass(Response):
    """
    Returned by a rule to signal that its conditions explicitly have **not**
    been met. In other words, the rule has all of the information it needs to
    determine that the system it's analyzing is not in the state the rule was
    meant to catch.

    An example rule might check whether a system is vulnerable to a well
    defined exploit or has a bug in a specific version of a package. If it can
    say for sure "the system does not have this exploit" or "the system does
    not have the buggy version of the package installed", then it should return
    an instance of :class:`make_pass`.

    Example:

    .. code-block:: python

        # completely made up package
        buggy = InstalledRpms.from_package("bash-3.4.23-1.el7")

        @rule(InstalledRpms)
        def report(installed_rpms):
           bash = installed_rpms.newest("bash")
           if bash == buggy:
               return make_fail("BASH_BUG_123", bash=bash)
           return make_pass("BASH", bash=bash)

    """
    response_type = "pass"
    key_name = "pass_key"


class make_info(Response):
    """
    Returned by a rule to surface information about a system.

    Example:

    .. code-block:: python

        @rule(InstalledRpms)
        def report(rpms):
           bash = rpms.newest("bash")
           return make_info("BASH_VERSION", bash=bash.nvra)

    """
    response_type = "info"
    key_name = "info_key"


class make_fingerprint(Response):
    response_type = "fingerprint"
    key_name = "fingerprint_key"


class make_metadata_key(Response):
    response_type = "metadata_key"
    key_name = "key"

    def __init__(self, key, value):
        super(make_metadata_key, self).__init__(key, value=value)

    def adjust_for_length(self, key, r, kwargs):
        return kwargs


class make_metadata(Response):
    """
    Allows a rule to convey addtional metadata about a system to downstream
    systems. It doesn't convey success or failure but purely information that
    may be aggregated with other make_metadata responses. As such, it has no
    response key.
    """
    response_type = "metadata"

    def __init__(self, **kwargs):
        super(make_metadata, self).__init__(None, **kwargs)


class _make_skip(Response):
    """
    Called automatically whenever a rule's dependencies aren't met. Likely to
    be deprecated or have its semantics changed. Do not call explicitly from
    rules.
    """
    response_type = "skip"

    def __str__(self):
        required = self.missing[0]
        at_least_one = self.missing[1]

        buf = StringIO()

        print("Missing Dependencies:", file=buf)

        if required:
            print("    Requires:", file=buf)
            for d in required:
                print("        %s" % dr.get_name(d), file=buf)
        if at_least_one:
            for alo in at_least_one:
                print("    At Least One Of:", file=buf)
                for d in alo:
                    print("        %s" % dr.get_name(d), file=buf)
        buf.seek(0)
        return buf.read()

    def __init__(self, rule_fqdn, missing):
        self.missing = missing
        details = dr.stringify_requirements(missing)
        super(_make_skip, self).__init__(None,
                                        rule_fqdn=rule_fqdn,
                                        reason="MISSING_REQUIREMENTS",
                                        details=details)


class make_none(Response):
    """
    Used to create a response for a rule that returns None

    This is not intended to be used by plugins, only infrastructure
    but it not private so that we can easily add it to reporting.
    """
    response_type = "none"
    key_name = "none_key"

    def __init__(self):
        super(make_none, self).__init__(key="NONE_KEY")



from insights.parsers import multipath_v4_ll
from insights.tests import context_wrap
import doctest


MULTIPATH_V4_LL_INFO = """
===== paths list =====
uuid hcil    dev dev_t pri dm_st chk_st vend/prod/rev       dev_st
     0:0:0:0 sda 8:0   -1  undef ready  VMware,Virtual disk running
     3:0:0:1 sdb 8:16  -1  undef ready  IET,VIRTUAL-DISK    running
     4:0:0:1 sdc 8:32  -1  undef ready  IET,VIRTUAL-DISK    running
Oct 28 14:02:44 | *word = 0, len = 1
Oct 28 14:02:44 | *word = E, len = 1
Oct 28 14:02:44 | *word = 1, len = 1
Oct 28 14:02:44 | *word = 0, len = 1
Oct 28 14:02:44 | *word = A, len = 1
Oct 28 14:02:44 | *word = 0, len = 1
mpathg (36f01faf000da360b0000033c528fea6d) dm-2 DELL,MD36xxi
size=54T features='3 queue_if_no_path pg_init_retries 50' hwhandler='1 rdac' wp=rw
|-+- policy='round-robin 0' prio=0 status=active
| |- 12:0:0:1 sdc 8:32   active ready running
| |- 11:0:0:1 sdi 8:128  active ready running
| |- 15:0:0:1 sdo 8:224  active ready running
| `- 17:0:0:1 sdv 65:80  active ready running
`-+- policy='round-robin 0' prio=0 status=enabled
  |- 13:0:0:1 sdf 8:80   active ready running
  |- 14:0:0:1 sdl 8:176  active ready running
  |- 16:0:0:1 sdr 65:16  active ready running
  `- 18:0:0:1 sdx 65:112 active ready running
mpathe (36f01faf000da3761000004323aa6fbce) dm-4 DELL,MD36xxi
size=44T features='3 queue_if_no_path pg_init_retries 55' hwhandler='2 rdac' wp=rx
|-+- policy='round-robin 0' prio=1 status=active
| |- 12:0:0:2 sdc 8:32   active ready running
| |- 11:0:0:2 sdi 8:128  active ready running
| |- 15:0:0:2 sdo 8:224  active ready running
| `- 17:0:0:2 sdv 65:80  active ready running
`-+- policy='round-robin 0' prio=1 status=enabled
  |- 13:0:0:2 sdf 8:80   active ready running
  |- 14:0:0:2 sdl 8:176  active ready running
  |- 16:0:0:2 sdr 65:16  active ready running
  `- 18:0:0:2 sdx 65:112 active ready running
36001405b1629f80d52a4c898f8856e43 dm-5 LIO-ORG ,block0_sdb
size=2.0G features='0' hwhandler='0' wp=rw
|-+- policy='service-time 0' prio=1 status=active
| `- 3:0:0:0 sdc 8:32 active ready running
`-+- policy='service-time 0' prio=1 status=enabled
  `- 4:0:0:0 sdb 8:16 active ready running
mpatha (1IET     00080001) dm-0 IET,VIRTUAL-DISK
size=16G features='0' hwhandler='0' wp=rw
|-+- policy='round-robin 0' prio=1 status=active
| `- 3:0:0:1 sdb 8:16 active ready running
`-+- policy='round-robin 0' prio=1 status=enabled
  `- 4:0:0:1 sdc 8:32 active ready running
1IET     00080001 dm-0 IET,VIRTUAL-DISK
size=16G features='0' hwhandler='0' wp=rw
|-+- policy='round-robin 0' prio=1 status=active
| `- 3:0:0:1 sdb 8:16 active ready running
`-+- policy='round-robin 0' prio=1 status=enabled
  `- 4:0:0:1 sdc 8:32 active ready running
mpathb (1IET     00080002) dm-8 COMPELNT,Compellent Vol
size=16G features='0' hwhandler='0' wp=rw
|-+- policy='round-robin 0' prio=1 status=active
| `- 3:0:0:1 sdb 8:16 active ready running
`-+- policy='round-robin 0' prio=1 status=enabled
  `- 4:0:0:1 sdc 8:32 active ready running
1IET     00080007 dm-19 COMPELNT,Compellent Vol
size=16G features='0' hwhandler='0' wp=rw
|-+- policy='round-robin 0' prio=1 status=active
| `- 3:0:0:1 sdb 8:16 active ready running
`-+- policy='round-robin 0' prio=1 status=enabled
  `- 4:0:0:1 sdc 8:32 active ready running
mpathc (test_with_no_devs) dm-1 uninitialized
size=10G features='0' hwhandler='0' wp=rw
""".strip()

MULTIPATH_V4_LL_INFO_RHEL_5 = """
sdz: size = 293601280
sdz: vendor = DGC
sdz: product = RAID 5
sdz: rev = 0430
sdz: h:b:t:l = 5:0:1:6
sdz: tgt_node_name = 0x50060160c4603569
sdz: path checker = emc_clariion (controller setting)
sdz: checker timeout = 60000 ms (sysfs setting)
sdz: state = 2
sr0: blacklisted
sr1: blacklisted
===== paths list =====
uuid hcil    dev  dev_t  pri dm_st  chk_st  vend/prod/rev
     5:0:1:7 sdaa 65:160 0   [undef][ready] DGC,RAID 5
     3:0:0:5 sdab 65:176 0   [undef][ready] DGC,RAID 5
     3:0:0:6 sdac 65:192 0   [undef][ready] DGC,RAID 5
     3:0:0:7 sdad 65:208 0   [undef][ready] DGC,RAID 5
     3:0:1:5 sdae 65:224 0   [undef][ready] DGC,RAID 5
     3:0:1:6 sdaf 65:240 0   [undef][ready] DGC,RAID 5
     3:0:1:7 sdag 66:0   0   [undef][ready] DGC,RAID 5
     0:2:0:0 sda  8:0    0   [undef][ready] DELL,PERC 6/i
     3:0:0:0 sdb  8:16   0   [undef][ready] DGC,RAID 5
     3:0:0:1 sdc  8:32   0   [undef][ready] DGC,RAID 5
     3:0:0:2 sdd  8:48   0   [undef][ready] DGC,RAID 5
     3:0:0:3 sde  8:64   0   [undef][ready] DGC,RAID 5
     3:0:0:4 sdf  8:80   0   [undef][ready] DGC,RAID 5
     3:0:1:0 sdg  8:96   0   [undef][ready] DGC,RAID 5
     3:0:1:1 sdh  8:112  0   [undef][ready] DGC,RAID 5
     3:0:1:2 sdi  8:128  0   [undef][ready] DGC,RAID 5
     3:0:1:3 sdj  8:144  0   [undef][ready] DGC,RAID 5
     3:0:1:4 sdk  8:160  0   [undef][ready] DGC,RAID 5
     5:0:0:0 sdl  8:176  0   [undef][ready] DGC,RAID 5
     5:0:0:1 sdm  8:192  0   [undef][ready] DGC,RAID 5
     5:0:0:2 sdn  8:208  0   [undef][ready] DGC,RAID 5
     5:0:0:3 sdo  8:224  0   [undef][ready] DGC,RAID 5
     5:0:0:4 sdp  8:240  0   [undef][ready] DGC,RAID 5
     5:0:1:0 sdq  65:0   0   [undef][ready] DGC,RAID 5
     5:0:1:1 sdr  65:16  0   [undef][ready] DGC,RAID 5
     5:0:1:2 sds  65:32  0   [undef][ready] DGC,RAID 5
     5:0:1:3 sdt  65:48  0   [undef][ready] DGC,RAID 5
     5:0:1:4 sdu  65:64  0   [undef][ready] DGC,RAID 5
     5:0:0:5 sdv  65:80  0   [undef][ready] DGC,RAID 5
     5:0:0:6 sdw  65:96  0   [undef][ready] DGC,RAID 5
     5:0:0:7 sdx  65:112 0   [undef][ready] DGC,RAID 5
     5:0:1:5 sdy  65:128 0   [undef][ready] DGC,RAID 5
     5:0:1:6 sdz  65:144 0   [undef][ready] DGC,RAID 5
params = 1 queue_if_no_path 1 emc 2 1 round-robin 0 2 1 8:160 1000 8:240 1000 round-robin 0 2 1 8:80 1000 65:64 1000
status = 2 0 1 0 2 1 A 0 2 0 8:160 A 0 8:240 A 0 E 0 2 0 8:80 A 0 65:64 A 0
*word = 1, len = 1
*word = queue_if_no_path, len = 16
*word = 1, len = 1
*word = emc, len = 3
sdu: getprio = /sbin/mpath_prio_emc /dev/%n (controller setting)
process 31154 forking to exec '/sbin/mpath_prio_emc /dev/sdu' ((nil))
forked 31158
sdu: prio = 0
*word = 2, len = 1
*word = 1, len = 1
L004 (360060160ade32800f2e3baf47665e211) dm-9 DGC,RAID 5
[size=100G][features=1 queue_if_no_path][hwhandler=1 emc][rw]
\_ round-robin 0 [prio=1][active]
 \_ 3:0:1:4 sdk  8:160  [active][ready]
 \_ 5:0:0:4 sdp  8:240  [active][ready]
\_ round-robin 0 [prio=0][enabled]
 \_ 3:0:0:4 sdf  8:80   [active][ready]
 \_ 5:0:1:4 sdu  65:64  [active][ready]
params = 1 queue_if_no_path 1 emc 2 1 round-robin 0 2 1 8:64 1000 65:48 1000 round-robin 0 2 1 8:144 1000 8:224 1000
status = 2 0 1 0 2 1 A 0 2 0 8:64 A 0 65:48 A 0 E 0 2 0 8:144 A 0 8:224 A 0
*word = 1, len = 1
*word = queue_if_no_path, len = 16
"""


def test_class_RHEL6():
    mp = multipath_v4_ll.MultipathDevices(context_wrap(MULTIPATH_V4_LL_INFO))
    assert len(mp.devices) == 7
    assert mp.devices[0] == {
        "alias": "mpathg",
        "wwid": "36f01faf000da360b0000033c528fea6d",
        "dm_name": "dm-2",
        "venprod": "DELL,MD36xxi",
        "size": "54T",
        "features": "3 queue_if_no_path pg_init_retries 50",
        "hwhandler": "1 rdac",
        "wp": "rw",
        "path_group": [{
            "policy": "round-robin 0",
            "prio": "0",
            "status": "active",
            "path": [
                ['12:0:0:1', 'sdc', '8:32', 'active', 'ready', 'running'],
                ['11:0:0:1', 'sdi', '8:128', 'active', 'ready', 'running'],
                ['15:0:0:1', 'sdo', '8:224', 'active', 'ready', 'running'],
                ['17:0:0:1', 'sdv', '65:80', 'active', 'ready', 'running']
            ]
        }, {
            "policy": "round-robin 0",
            "prio": "0",
            "status": "enabled",
            "path": [
                ['13:0:0:1', 'sdf', '8:80', 'active', 'ready', 'running'],
                ['14:0:0:1', 'sdl', '8:176', 'active', 'ready', 'running'],
                ['16:0:0:1', 'sdr', '65:16', 'active', 'ready', 'running'],
                ['18:0:0:1', 'sdx', '65:112', 'active', 'ready', 'running']
            ]
        }]
    }
    assert mp.devices[0].get('size') == '54T'
    assert mp.devices[1].get('path_group') == [{
        "policy": "round-robin 0",
        "prio": "1",
        "status": "active",
        "path": [
            ['12:0:0:2', 'sdc', '8:32', 'active', 'ready', 'running'],
            ['11:0:0:2', 'sdi', '8:128', 'active', 'ready', 'running'],
            ['15:0:0:2', 'sdo', '8:224', 'active', 'ready', 'running'],
            ['17:0:0:2', 'sdv', '65:80', 'active', 'ready', 'running']
        ]
    }, {
        "policy": "round-robin 0",
        "prio": "1",
        "status": "enabled",
        "path": [
            ['13:0:0:2', 'sdf', '8:80', 'active', 'ready', 'running'],
            ['14:0:0:2', 'sdl', '8:176', 'active', 'ready', 'running'],
            ['16:0:0:2', 'sdr', '65:16', 'active', 'ready', 'running'],
            ['18:0:0:2', 'sdx', '65:112', 'active', 'ready', 'running']
        ]
    }]
    assert mp.devices[2].get('hwhandler') == "0"
    assert mp.devices[3].get('alias') == "mpatha"
    assert mp.devices[4].get('wwid') == "1IET     00080001"
    assert mp.devices[5].get('venprod') == "COMPELNT,Compellent Vol"
    assert mp.devices[5].get('dm_name') == "dm-8"
    assert mp.devices[6].get('venprod') == "COMPELNT,Compellent Vol"
    assert mp.devices[6].get('dm_name') == "dm-19"
    # Note that there's no data for the made-up 'mpathc', since there's no
    # path group information and only devices with path group information
    # get saved.
    assert mp.dms == ['dm-2', 'dm-4', 'dm-5', 'dm-0', 'dm-0', 'dm-8', 'dm-19']
    assert mp.by_dm['dm-2'] == mp.devices[0]
    assert mp.aliases == ['mpathg', 'mpathe', 'mpatha', 'mpathb']
    assert mp.by_alias['mpathg'] == mp.devices[0]
    assert mp.wwids == [
        '36f01faf000da360b0000033c528fea6d',
        '36f01faf000da3761000004323aa6fbce',
        '36001405b1629f80d52a4c898f8856e43',
        '1IET     00080001',
        '1IET     00080001',
        '1IET     00080002',
        '1IET     00080007'
    ]
    assert mp.by_wwid['1IET     00080001'] == mp.devices[4]

    # Pseudo list accessors
    assert len(mp) == 7
    for i, item in enumerate(mp):
        assert item == mp.devices[i]
        assert item == mp[i]

    assert len(mp.raw_info_lines) == 11
    assert "===== paths list =====" in mp.raw_info_lines


MULTIPATH_V4_LL_DOC = """
===== paths list =====
uuid hcil    dev dev_t pri dm_st chk_st vend/prod/rev       dev_st
     0:0:0:0 sda 8:0   -1  undef ready  VMware,Virtual disk running
     3:0:0:1 sdb 8:16  -1  undef ready  IET,VIRTUAL-DISK    running
     4:0:0:1 sdc 8:32  -1  undef ready  IET,VIRTUAL-DISK    running
Oct 28 14:02:44 | *word = 0, len = 1
Oct 28 14:02:44 | *word = E, len = 1
Oct 28 14:02:44 | *word = 1, len = 1
Oct 28 14:02:44 | *word = 0, len = 1
Oct 28 14:02:44 | *word = A, len = 1
Oct 28 14:02:44 | *word = 0, len = 1
mpathg (36f01faf000da360b0000033c528fea6d) dm-2 DELL,MD36xxi
size=54T features='3 queue_if_no_path pg_init_retries 50' hwhandler='1 rdac' wp=rw
|-+- policy='round-robin 0' prio=0 status=active
| |- 12:0:0:1 sdc 8:32   active ready running
| |- 11:0:0:1 sdi 8:128  active ready running
| |- 15:0:0:1 sdo 8:224  active ready running
| `- 17:0:0:1 sdv 65:80  active ready running
`-+- policy='round-robin 0' prio=0 status=enabled
  |- 13:0:0:1 sdf 8:80   active ready running
  |- 14:0:0:1 sdl 8:176  active ready running
  |- 16:0:0:1 sdr 65:16  active ready running
  `- 18:0:0:1 sdx 65:112 active ready running
mpathe (36f01faf000da3761000004323aa6fbce) dm-4 DELL,MD36xxi
size=54T features='3 queue_if_no_path pg_init_retries 55' hwhandler='1 rdac' wp=rw
|-+- policy='round-robin 0' prio=0 status=active
| |- 13:0:0:2 sdg 8:96   active faulty running
| |- 14:0:0:2 sdm 8:192  active faulty running
| |- 16:0:0:2 sds 65:32  active faulty running
| `- 18:0:0:2 sdy 65:128 active faulty running
`-+- policy='round-robin 0' prio=0 status=enabled
  |- 12:0:0:2 sdd 8:48   active faulty running
  |- 11:0:0:2 sdj 8:144  active faulty running
  |- 15:0:0:2 sdp 8:240  active faulty running
  `- 17:0:0:2 sdw 65:96  active faulty running
36001405b1629f80d52a4c898f8856e43 dm-5 LIO-ORG ,block0_sdb
size=2.0G features='0' hwhandler='0' wp=rw
|-+- policy='service-time 0' prio=1 status=active
| `- 3:0:0:0 sdc 8:32 active ready running
`-+- policy='service-time 0' prio=1 status=enabled
  `- 4:0:0:0 sdb 8:16 active ready running
"""


def test_doc_examples():
    env = {
            'MultipathDevices': multipath_v4_ll.MultipathDevices,
            'mpaths': multipath_v4_ll.MultipathDevices(context_wrap(MULTIPATH_V4_LL_DOC)),
          }
    failed, total = doctest.testmod(multipath_v4_ll, globs=env)
    assert failed == 0



"""
LsIPAIdoverrideMemberof - command ``ls -lan /usr/share/ipa/ui/js/plugins/idoverride-memberof``
==============================================================================================

The ``ls -lan /usr/share/ipa/ui/js/plugins/idoverride-memberof`` command provides information for the listing of the
``/usr/share/ipa/ui/js/plugins/idoverride-memberof`` files. See the ``FileListing`` class for a more complete description
of the available features of the class.

Sample ``ls -lan /usr/share/ipa/ui/js/plugins/idoverride-memberof`` output::

    /usr/share/ipa/ui/js/plugins/idoverride-memberof:
    total 0
    drwxr-xr-x. 2 0 0 0 Nov 11 11:44 .
    drwxr-xr-x. 4 0 0 0 Nov 11 11:44 ..
    -rw-rw-r--. 1 0 0 0 Nov 11 11:44 idoverride-memberof.js
    -rw-rw-r--. 1 0 0 0 Nov 11 11:44 idoverride-admemberof.js

Examples:
    >>> '/usr/share/ipa/ui/js/plugins/idoverride-memberof' in ls_ipa_idoverride_memberof
    True
    >>> ls_ipa_idoverride_memberof.files_of('/usr/share/ipa/ui/js/plugins/idoverride-memberof') == ['idoverride-memberof.js', 'idoverride-admemberof.js']
    True
"""

from insights.specs import Specs
from .. import parser, CommandParser, FileListing
from insights.util import deprecated


@parser(Specs.ls_ipa_idoverride_memberof)
class LsIPAIdoverrideMemberof(CommandParser, FileListing):
    """
    .. warning::
        This class is deprecated and will be removed from 3.5.0.
        Please use the :class:`insights.parsers.ls.LSlan` instead.

    Parses output of ``ls -lan /usr/share/ipa/ui/js/plugins/idoverride-memberof`` command.
    """
    def __init__(self, *args, **kwargs):
        deprecated(LsIPAIdoverrideMemberof, "Please use the :class:`insights.parsers.ls.LSlan` instead.", "3.5.0")
        super(LsIPAIdoverrideMemberof, self).__init__(*args, **kwargs)



import json
import pytest

try:
    from unittest.mock import patch
    builtin_open = "builtins.open"
except Exception:
    from mock import patch
    builtin_open = "__builtin__.open"

from os import path

from insights.core.exceptions import SkipComponent, ContentException
from insights.specs.datasources.leapp import leapp_report, migration_results


with open(path.join(path.dirname(__file__), 'leapp-report.json'), 'r') as fp:
    LEAPP_REPORT_OK = json.load(fp)

with open(path.join(path.dirname(__file__), 'leapp-report.ret'), 'r') as fp:
    LEAPP_REPORT_RESULT = fp.readlines()

LEAPP_REPORT_NG_1 = json.loads("""
{ "entries":[
  {
    "a": "B"
  }
 ]
}
""".strip())

LEAPP_REPORT_NG_2 = MIGRATION_RESULTS_NG_2 = json.loads("{}")

MIGRATION_RESULTS_OK = json.loads("""
{
   "activities": [
     {
       "activity": "preupgrade",
       "activity_ended": "2023-08-22T08:56:26.971009Z",
       "activity_started": "2023-08-22T08:55:47.798008Z",
       "env": {
         "LEAPP_COMMON_FILES": ":/etc/leapp/repos.d/system_upgrade/common/files:/etc/leapp/repos.d/system_upgrade/el8toel9/files",
         "LEAPP_COMMON_TOOLS": ":/etc/leapp/repos.d/system_upgrade/common/tools:/etc/leapp/repos.d/system_upgrade/el8toel9/tools",
         "LEAPP_CURRENT_PHASE": "FactsCollection",
         "LEAPP_DEBUG": "0",
         "LEAPP_EXECUTION_ID": "1edff870-626d-41ba-854c-8f9dc8f20dc3",
         "LEAPP_EXPERIMENTAL": "0",
         "LEAPP_HOSTNAME": "kvm-04-guest17.lab.eng.rdu2.redhat.com",
         "LEAPP_IPU_IN_PROGRESS": "8to9",
         "LEAPP_NO_RHSM": "0",
         "LEAPP_UNSUPPORTED": "0",
         "LEAPP_UPGRADE_PATH_FLAVOUR": "default",
         "LEAPP_UPGRADE_PATH_TARGET_RELEASE": "9.0",
         "LEAPP_VERBOSE": "0"
       },
       "executed": "/usr/bin/leapp preupgrade",
       "packages": [
         {
           "nevra": "leapp-0.15.1-1.el8_6.2.noarch",
           "signature": "RSA/SHA256, Mon 13 Mar 2023 12:08:57 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "python3-leapp-0.15.1-1.el8_6.2.noarch",
           "signature": "RSA/SHA256, Mon 13 Mar 2023 12:08:58 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "leapp-upgrade-el8toel9-0.18.0-1.el8_6.3.noarch",
           "signature": "RSA/SHA256, Mon 05 Jun 2023 12:54:10 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "leapp-deps-0.15.1-1.el8_6.2.noarch",
           "signature": "RSA/SHA256, Mon 13 Mar 2023 12:08:58 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "leapp-upgrade-el8toel9-deps-0.18.0-1.el8_6.3.noarch",
           "signature": "RSA/SHA256, Mon 05 Jun 2023 12:54:11 PM EDT, Key ID 199e2f91fd431d51"
         }
       ],
       "run_id": "1edff870-626d-41ba-854c-8f9dc8f20dc3",
       "source_os": "Red Hat Enterprise Linux 8.6",
       "success": false,
       "target_os": "Red Hat Enterprise Linux 9.0",
       "version": "0.15.1"
     },
     {
       "activity": "upgrade",
       "activity_ended": "2023-08-22T09:21:40.483395Z",
       "activity_started": "2023-08-22T09:13:50.876865Z",
       "env": {
         "LEAPP_COMMON_FILES": ":/etc/leapp/repos.d/system_upgrade/common/files:/etc/leapp/repos.d/system_upgrade/el8toel9/files",
         "LEAPP_COMMON_TOOLS": ":/etc/leapp/repos.d/system_upgrade/common/tools:/etc/leapp/repos.d/system_upgrade/el8toel9/tools",
         "LEAPP_CURRENT_PHASE": "InterimPreparation",
         "LEAPP_DEBUG": "0",
         "LEAPP_EXECUTION_ID": "b3baa3f4-de86-4c4b-a7f3-bbe271207efd",
         "LEAPP_EXPERIMENTAL": "0",
         "LEAPP_HOSTNAME": "kvm-04-guest17.lab.eng.rdu2.redhat.com",
         "LEAPP_IPU_IN_PROGRESS": "8to9",
         "LEAPP_NO_RHSM": "0",
         "LEAPP_UNSUPPORTED": "0",
         "LEAPP_UPGRADE_PATH_FLAVOUR": "default",
         "LEAPP_UPGRADE_PATH_TARGET_RELEASE": "9.0",
         "LEAPP_VERBOSE": "0"
       },
       "executed": "/usr/bin/leapp upgrade",
       "packages": [
         {
           "nevra": "leapp-0.15.1-1.el8_6.2.noarch",
           "signature": "RSA/SHA256, Mon 13 Mar 2023 12:08:57 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "python3-leapp-0.15.1-1.el8_6.2.noarch",
           "signature": "RSA/SHA256, Mon 13 Mar 2023 12:08:58 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "leapp-upgrade-el8toel9-0.18.0-1.el8_6.3.noarch",
           "signature": "RSA/SHA256, Mon 05 Jun 2023 12:54:10 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "leapp-deps-0.15.1-1.el8_6.2.noarch",
           "signature": "RSA/SHA256, Mon 13 Mar 2023 12:08:58 PM EDT, Key ID 199e2f91fd431d51"
         },
         {
           "nevra": "leapp-upgrade-el8toel9-deps-0.18.0-1.el8_6.3.noarch",
           "signature": "RSA/SHA256, Mon 05 Jun 2023 12:54:11 PM EDT, Key ID 199e2f91fd431d51"
         }
       ],
       "run_id": "b3baa3f4-de86-4c4b-a7f3-bbe271207efd",
       "source_os": "Red Hat Enterprise Linux 8.6",
       "success": true,
       "target_os": "Red Hat Enterprise Linux 9.0",
       "version": "0.15.1"
     }
  ]
}""".strip())

MIGRATION_RESULTS_RET = """
[
  {
    "activity": "preupgrade",
    "activity_ended": "2023-08-22T08:56:26.971009Z",
    "activity_started": "2023-08-22T08:55:47.798008Z",
    "env": {
      "LEAPP_CURRENT_PHASE": "FactsCollection",
      "LEAPP_EXPERIMENTAL": "0",
      "LEAPP_NO_RHSM": "0",
      "LEAPP_UNSUPPORTED": "0"
    },
    "run_id": "1edff870-626d-41ba-854c-8f9dc8f20dc3",
    "source_os": "Red Hat Enterprise Linux 8.6",
    "success": false,
    "target_os": "Red Hat Enterprise Linux 9.0",
    "version": "0.15.1"
  },
  {
    "activity": "upgrade",
    "activity_ended": "2023-08-22T09:21:40.483395Z",
    "activity_started": "2023-08-22T09:13:50.876865Z",
    "env": {
      "LEAPP_CURRENT_PHASE": "InterimPreparation",
      "LEAPP_EXPERIMENTAL": "0",
      "LEAPP_NO_RHSM": "0",
      "LEAPP_UNSUPPORTED": "0"
    },
    "run_id": "b3baa3f4-de86-4c4b-a7f3-bbe271207efd",
    "source_os": "Red Hat Enterprise Linux 8.6",
    "success": true,
    "target_os": "Red Hat Enterprise Linux 9.0",
    "version": "0.15.1"
  }
]
""".strip()

MIGRATION_RESULTS_NG_1 = json.loads("""
{ "activities":[
  {
    "a": "B"
  }
 ]
}
""".strip())


@patch("json.load", return_value=LEAPP_REPORT_OK)
@patch("os.path.isfile", return_value=True)
@patch(builtin_open)
def test_leapp_report_ok(m_open, m_isfile, m_load):
    result = leapp_report({})
    result_json = json.loads(''.join(result.content).strip())
    expected_json = json.loads(''.join(LEAPP_REPORT_RESULT).strip())
    for ret in result_json:
        assert ret in expected_json


@patch("json.load", return_value=LEAPP_REPORT_NG_1)
@patch("os.path.isfile", return_value=True)
@patch(builtin_open)
def test_leapp_report_nothing(m_open, m_isfile, m_load):
    with pytest.raises(SkipComponent) as ce:
        leapp_report({})
    assert "Nothing" in str(ce)


@patch("json.load", return_value=LEAPP_REPORT_NG_2)
@patch("os.path.isfile", return_value=True)
@patch(builtin_open)
def test_leapp_report_ng_2(m_open, m_isfile, m_load):
    with pytest.raises(ContentException) as ce:
        leapp_report({})
    assert "Nothing" in str(ce)


@patch("os.path.isfile", return_value=False)
def test_leapp_report_no_file(isfile):
    with pytest.raises(SkipComponent):
        leapp_report({})


@patch("json.load", return_value=MIGRATION_RESULTS_OK)
@patch("os.path.isfile", return_value=True)
@patch(builtin_open)
def test_leapp_migration_results_ok(m_open, m_isfile, m_load):
    result = migration_results({})
    result_json = json.loads(''.join(result.content).strip())
    expected_json = json.loads(''.join(MIGRATION_RESULTS_RET).strip())
    for ret in result_json:
        assert ret in expected_json


@patch("json.load", return_value=MIGRATION_RESULTS_NG_1)
@patch("os.path.isfile", return_value=True)
@patch(builtin_open)
def test_leapp_migration_results_nothing(m_open, m_isfile, m_load):
    with pytest.raises(SkipComponent) as ce:
        migration_results({})
    assert "Nothing" in str(ce)


@patch("json.load", return_value=MIGRATION_RESULTS_NG_2)
@patch("os.path.isfile", return_value=True)
@patch(builtin_open)
def test_leapp_migration_results_ng_2(m_open, m_isfile, m_load):
    with pytest.raises(ContentException) as ce:
        migration_results({})
    assert "Nothing" in str(ce)


@patch("os.path.isfile", return_value=False)
def test_leapp_migration_results_no_file(isfile):
    with pytest.raises(SkipComponent):
        migration_results({})



import pytest

from mock.mock import patch

from insights.core.context import HostContext
from insights.core.exceptions import SkipComponent
from insights.parsers.docker_list import DockerListContainers
from insights.parsers.podman_list import PodmanListContainers
from insights.specs.datasources.container import running_rhel_containers
from insights.tests import context_wrap

PODMAN_LIST_CONTAINERS_2_UP = """
CONTAINER ID                                                       IMAGE         COMMAND                                            CREATED             STATUS                      PORTS                  NAMES               SIZE
03e2861336a76e29155836113ff6560cb70780c32f95062642993b2b3d0fc216   rhel7_httpd   "/usr/sbin/httpd -DFOREGROUND"                     45 seconds ago      Up 37 seconds               0.0.0.0:8080->80/tcp   angry_saha          796 B (virtual 669.2 MB)
05516ea08b565e37e2a4bca3333af40a240c368131b77276da8dec629b7fe102   bd8638c869ea  "/bin/sh -c 'yum install -y vsftpd-2.2.2-6.el6'"   18 hours ago        Up 18 hours ago                                    tender_rosalind     4.751 MB (virtual 200.4 MB)
""".strip()

DOCKER_LIST_CONTAINERS_1_UP = """
CONTAINER ID                                                       IMAGE         COMMAND                                            CREATED             STATUS                      PORTS                  NAMES               SIZE
d3e2861336a76e29155836113ff6560cb70780c32f95062642993b2b3d0fc216   rhel7_httpd   "/usr/sbin/httpd -DFOREGROUND"                     45 seconds ago      Up 37 seconds               0.0.0.0:8080->80/tcp   angry_saha          796 B (virtual 669.2 MB)
d5516ea08b565e37e2a4bca3333af40a240c368131b77276da8dec629b7fe102   bd8638c869ea  "/bin/sh -c 'yum install -y vsftpd-2.2.2-6.el6'"   18 hours ago        Exited (137) 18 hours ago                          tender_rosalind     4.751 MB (virtual 200.4 MB)
""".strip()

FEDORA = """
Fedora release 23 (Twenty Three)
""".strip()

REDHAT_RELEASE7 = """
Red Hat Enterprise Linux release 7.3
""".strip()


def fake_shell_out(cmd, split=True, timeout=None, keep_rc=False, env=None, signum=None):
    tmp_cmd = cmd.strip().split()
    if 'podman' in tmp_cmd[0]:
        return [REDHAT_RELEASE7, ]
    if 'docker' in tmp_cmd[0]:
        return [FEDORA, ]
    raise Exception()


@patch("insights.core.context.HostContext.shell_out", return_value=[REDHAT_RELEASE7, ])
def test_get_running_rhel_containers_both_ok(fso):
    p_ctn = PodmanListContainers(context_wrap(PODMAN_LIST_CONTAINERS_2_UP))
    d_ctn = DockerListContainers(context_wrap(DOCKER_LIST_CONTAINERS_1_UP))
    assert p_ctn is not None
    assert d_ctn is not None

    broker = {
        PodmanListContainers: p_ctn,
        DockerListContainers: d_ctn,
        HostContext: HostContext()}

    ret = running_rhel_containers(broker)
    assert len(ret) == 3
    assert ('rhel7_httpd', 'podman', '03e2861336a7') in ret
    assert ('bd8638c869ea', 'podman', '05516ea08b56') in ret
    assert ('rhel7_httpd', 'docker', 'd3e2861336a7') in ret
    # the stopped container is not collected


@patch("insights.core.context.HostContext.shell_out", side_effect=fake_shell_out)
def test_get_running_rhel_containers_podman_only(fso):
    p_ctn = PodmanListContainers(context_wrap(PODMAN_LIST_CONTAINERS_2_UP))
    d_ctn = DockerListContainers(context_wrap(DOCKER_LIST_CONTAINERS_1_UP))
    assert p_ctn is not None
    assert d_ctn is not None

    broker = {
        PodmanListContainers: p_ctn,
        DockerListContainers: d_ctn,
        HostContext: HostContext()}

    ret = running_rhel_containers(broker)
    assert len(ret) == 2
    assert ('rhel7_httpd', 'podman', '03e2861336a7') in ret
    assert ('bd8638c869ea', 'podman', '05516ea08b56') in ret
    # docker container is from Fedora image, not collected


@patch("insights.core.context.HostContext.shell_out", return_value=[REDHAT_RELEASE7, ])
def test_get_running_rhel_containers_skip_dup(fso):
    p_ctn = PodmanListContainers(context_wrap(PODMAN_LIST_CONTAINERS_2_UP))
    # use the 'podman list' result as input for docker
    d_ctn = DockerListContainers(context_wrap(PODMAN_LIST_CONTAINERS_2_UP))
    assert p_ctn is not None
    assert d_ctn is not None

    broker = {
        PodmanListContainers: p_ctn,
        DockerListContainers: d_ctn,
        HostContext: HostContext()}

    ret = running_rhel_containers(broker)
    assert len(ret) == 2
    assert ('rhel7_httpd', 'podman', '03e2861336a7') in ret
    assert ('bd8638c869ea', 'podman', '05516ea08b56') in ret
    # duplicated container is removed from docker, not collected


@patch("insights.core.context.HostContext.shell_out", return_value=[FEDORA, ])
def test_get_running_rhel_containers_empty(fso):
    p_ctn = PodmanListContainers(context_wrap(PODMAN_LIST_CONTAINERS_2_UP))
    d_ctn = DockerListContainers(context_wrap(DOCKER_LIST_CONTAINERS_1_UP))
    assert p_ctn is not None
    assert d_ctn is not None

    broker = {
        PodmanListContainers: p_ctn,
        DockerListContainers: d_ctn,
        HostContext: HostContext()}

    with pytest.raises(SkipComponent):
        ret = running_rhel_containers(broker)
        assert len(ret) == 0



"""
NeutronMetadataAgentIni - file ``/etc/neutron/metadata_agent.ini``
==================================================================
"""
from insights.core import IniConfigFile
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.specs import Specs

add_filter(Specs.neutron_metadata_agent_ini, ["["])


@parser(Specs.neutron_metadata_agent_ini)
class NeutronMetadataAgentIni(IniConfigFile):
    """
    Parse the ``/etc/neutron/metadata_agent.ini`` configuration file.

    Sample configuration::

        [DEFAULT]
        debug = False
        auth_url = http://localhost:35357/v2.0
        auth_insecure = False
        admin_tenant_name = service
        admin_user = neutron
        nova_metadata_ip = 127.0.0.1
        nova_metadata_port = 8775
        nova_metadata_protocol = http
        metadata_workers =0
        metadata_backlog = 4096

        [AGENT]
        log_agent_heartbeats = False

    Examples:
        >>> type(metadata_agent_ini)
        <class 'insights.parsers.neutron_metadata_agent_conf.NeutronMetadataAgentIni'>
        >>> metadata_agent_ini.has_option('AGENT', 'log_agent_heartbeats')
        True
        >>> metadata_agent_ini.get("DEFAULT", "auth_url") == 'http://localhost:35357/v2.0'
        True
        >>> metadata_agent_ini.getint("DEFAULT", "metadata_backlog")
        4096
    """
    pass



"""
nova_log - files ``/var/log/nova/*.log``
========================================
This module contains classes to parse logs under ``/var/log/nova/``
"""
from .. import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.nova_api_log)
class NovaApiLog(LogFileOutput):
    """Class for parsing the ``/var/log/nova/nova-api.log`` file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`
    """
    pass


@parser(Specs.nova_compute_log)
class NovaComputeLog(LogFileOutput):
    """Class for parsing the ``/var/log/nova/nova-compute.log`` file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`
    """
    pass



"""
Collect all the interesting data for analysis
"""
from __future__ import absolute_import
import os
import errno
import json
import logging
import copy
import glob
import six
import shlex
import re
from itertools import chain
from subprocess import Popen, PIPE, STDOUT
from tempfile import NamedTemporaryFile

from insights.core.blacklist import BLACKLISTED_SPECS
from insights.util import mangle
from ..contrib.soscleaner import SOSCleaner
from .utilities import _expand_paths, get_version_info, systemd_notify_init_thread, get_tags
from .constants import InsightsConstants as constants
from .insights_spec import InsightsFile, InsightsCommand
from .archive import InsightsArchive

APP_NAME = constants.app_name
logger = logging.getLogger(__name__)
# python 2.7
SOSCLEANER_LOGGER = logging.getLogger('soscleaner')
SOSCLEANER_LOGGER.setLevel(logging.ERROR)
# python 2.6
SOSCLEANER_LOGGER = logging.getLogger('insights-client.soscleaner')
SOSCLEANER_LOGGER.setLevel(logging.ERROR)


def _process_content_redaction(filepath, exclude, regex=False):
    '''
    Redact content from a file, based on
    /etc/insights-client/.exp.sed and and the contents of "exclude"

    filepath    file to modify
    exclude     list of strings to redact
    regex       whether exclude is a list of regular expressions

    Returns the file contents with the specified data removed
    '''
    logger.debug('Processing %s...', filepath)

    # password removal
    sedcmd = Popen(['sed', '-rf', constants.default_sed_file, filepath], stdout=PIPE)
    # patterns removal
    if exclude:
        exclude_file = NamedTemporaryFile()
        exclude_file.write("\n".join(exclude).encode('utf-8'))
        exclude_file.flush()
        if regex:
            flag = '-E'
        else:
            flag = '-F'
        grepcmd = Popen(['grep', '-v', flag, '-f', exclude_file.name], stdin=sedcmd.stdout, stdout=PIPE)
        sedcmd.stdout.close()
        stdout, stderr = grepcmd.communicate()
        logger.debug('Process status: %s', grepcmd.returncode)
    else:
        stdout, stderr = sedcmd.communicate()
        logger.debug('Process status: %s', sedcmd.returncode)
    logger.debug('Process stderr: %s', stderr)
    return stdout


class DataCollector(object):
    '''
    Run commands and collect files
    '''

    def __init__(self, config, archive_=None, mountpoint=None):
        self.config = config
        self.archive = archive_ if archive_ else InsightsArchive(config)
        self.mountpoint = '/'
        if mountpoint:
            self.mountpoint = mountpoint
        self.hostname_path = None

    def _write_branch_info(self, branch_info):
        logger.debug("Writing branch information to archive...")
        self.archive.add_metadata_to_archive(
            json.dumps(branch_info), '/branch_info')

    def _write_display_name(self):
        if self.config.display_name:
            logger.debug("Writing display_name to archive...")
            self.archive.add_metadata_to_archive(
                self.config.display_name, '/display_name')

    def _write_ansible_host(self):
        if self.config.ansible_host:
            logger.debug("Writing ansible_host to archive...")
            self.archive.add_metadata_to_archive(
                self.config.ansible_host, '/ansible_host')

    def _write_version_info(self):
        logger.debug("Writing version information to archive...")
        version_info = get_version_info()
        self.archive.add_metadata_to_archive(
            json.dumps(version_info), '/version_info')

    def _write_tags(self):
        tags = get_tags()
        # NOTE:
        # The following code is also used by datasource 'tags'
        # - insights.specs.datasources.tags
        # Please keep them consistence before removing this.
        if tags is not None:
            def f(k, v):
                if type(v) is list:
                    col = []
                    for val in v:
                        col.append(f(k, val))
                    return list(chain.from_iterable(col))
                elif type(v) is dict:
                    col = []
                    for key, val in v.items():
                        col.append(f(k + ":" + key, val))
                    return list(chain.from_iterable(col))
                else:
                    return [{"key": k, "value": v, "namespace": constants.app_name}]
            logger.debug("Writing tags to archive...")
            t = []
            for k, v in tags.items():
                iv = f(k, v)
                t.append(iv)
            t = list(chain.from_iterable(t))
            self.archive.add_metadata_to_archive(json.dumps(t), '/tags.json')

    def _write_blacklist_report(self, blacklist_report):
        logger.debug("Writing blacklist report to archive...")
        self.archive.add_metadata_to_archive(
            json.dumps(blacklist_report), '/blacklist_report')

    def _write_blacklisted_specs(self):
        logger.debug("Writing blacklisted specs to archive...")

        if BLACKLISTED_SPECS:
            self.archive.add_metadata_to_archive(
                json.dumps({"specs": BLACKLISTED_SPECS}), '/blacklisted_specs')

    def _write_egg_release(self):
        logger.debug("Writing egg release to archive...")
        egg_release = ''
        try:
            with open(constants.egg_release_file) as fil:
                egg_release = fil.read()
        except (IOError, MemoryError) as e:
            logger.debug('Could not read the egg release file: %s', str(e))
        try:
            os.remove(constants.egg_release_file)
        except OSError as e:
            logger.debug('Could not remove the egg release file: %s', str(e))

        try:
            self.archive.add_metadata_to_archive(
                egg_release, '/egg_release')
        except OSError as e:
            logger.debug('Could not add the egg release file to the archive: %s', str(e))
            self.archive.add_metadata_to_archive(
                '', '/egg_release')

    def _write_collection_stats(self, collection_stats):
        logger.debug("Writing collection stats to archive...")
        self.archive.add_metadata_to_archive(
            json.dumps(collection_stats), '/collection_stats')

    def _write_rhsm_facts(self, hashed_fqdn, ip_csv):
        logger.info('Writing RHSM facts to %s...', constants.rhsm_facts_file)
        ips_list = ''
        with open(ip_csv) as fil:
            # create IP list as JSON block with format
            # [
            #   {
            #     original: <original IP>
            #     obfuscated: <obfuscated IP>
            #   }
            # ]

            ips_list = fil.readlines()
            headings = ips_list[0].strip().split(',')
            # set the indices for the IPs
            if 'original' in headings[0].lower():
                # soscleaner 0.4.4, original first
                org = 0
                obf = 1
            else:
                # soscleaner 0.2.2, obfuscated first
                org = 1
                obf = 0

            ip_block = []
            for line in ips_list[1:]:
                ipset = line.strip().split(',')
                ip_block.append(
                    {
                        'original': ipset[org],
                        'obfuscated': ipset[obf]
                    })

        facts = {
            'insights_client.obfuscate_hostname_enabled': self.config.obfuscate_hostname,
            'insights_client.hostname': hashed_fqdn,
            'insights_client.obfuscate_ip_enabled': self.config.obfuscate,
            'insights_client.ips': json.dumps(ip_block)
        }

        try:
            with open(constants.rhsm_facts_file, 'w') as fil:
                json.dump(facts, fil)
        except (IOError, OSError) as e:
            logger.error('Could not write to %s: %s', constants.rhsm_facts_file, str(e))

    def _run_pre_command(self, pre_cmd):
        '''
        Run a pre command to get external args for a command
        '''
        logger.debug('Executing pre-command: %s', pre_cmd)
        try:
            pre_proc = Popen(pre_cmd, stdout=PIPE, stderr=STDOUT, shell=True)
        except OSError as err:
            if err.errno == errno.ENOENT:
                logger.debug('Command %s not found', pre_cmd)
            return
        stdout, stderr = pre_proc.communicate()
        the_return_code = pre_proc.poll()
        logger.debug("Pre-command results:")
        logger.debug("STDOUT: %s", stdout)
        logger.debug("STDERR: %s", stderr)
        logger.debug("Return Code: %s", the_return_code)
        if the_return_code != 0:
            return []
        if six.PY3:
            stdout = stdout.decode('utf-8')
        return stdout.splitlines()

    def _parse_file_spec(self, spec):
        '''
        Separate wildcard specs into more specs
        '''
        # separate wildcard specs into more specs
        if '*' in spec['file']:
            expanded_paths = _expand_paths(spec['file'])
            if not expanded_paths:
                return []
            expanded_specs = []
            for p in expanded_paths:
                _spec = copy.copy(spec)
                _spec['file'] = p
                expanded_specs.append(_spec)
            return expanded_specs

        else:
            return [spec]

    def _parse_glob_spec(self, spec):
        '''
        Grab globs of things
        '''
        some_globs = glob.glob(spec['glob'])
        if not some_globs:
            return []
        el_globs = []
        for g in some_globs:
            _spec = copy.copy(spec)
            _spec['file'] = g
            el_globs.append(_spec)
        return el_globs

    def _blacklist_check(self, cmd):
        def _get_nested_parts(cmd):
            parts = shlex.split(cmd.replace(';', ' '))
            all_parts = parts[:]
            for p in parts:
                if len(shlex.split(p)) > 1:
                    all_parts += _get_nested_parts(p)
            return all_parts

        cmd_parts = _get_nested_parts(cmd)
        return len(set.intersection(set(cmd_parts),
                   constants.command_blacklist)) > 0

    def _parse_command_spec(self, spec, precmds):
        '''
        Run pre_commands
        '''
        if 'pre_command' in spec:
            precmd_alias = spec['pre_command']
            try:
                precmd = precmds[precmd_alias]

                if self._blacklist_check(precmd):
                    raise RuntimeError("Command Blacklist: " + precmd)

                args = self._run_pre_command(precmd)
                logger.debug('Pre-command results: %s', args)

                expanded_specs = []
                for arg in args:
                    _spec = copy.copy(spec)
                    _spec['command'] = _spec['command'] + ' ' + arg
                    expanded_specs.append(_spec)
                return expanded_specs
            except LookupError:
                logger.debug('Pre-command %s not found. Skipping %s...',
                             precmd_alias, spec['command'])
                return []
        else:
            return [spec]

    def run_collection(self, conf, rm_conf, branch_info, blacklist_report):
        '''
        Run specs and collect all the data
        '''
        # initialize systemd-notify thread
        systemd_notify_init_thread()

        self.archive.create_archive_dir()
        self.archive.create_command_dir()

        collection_stats = {}

        if rm_conf is None:
            rm_conf = {}
        logger.debug('Beginning to run collection spec...')

        rm_commands = rm_conf.get('commands', [])
        rm_files = rm_conf.get('files', [])

        for c in conf['commands']:
            # remember hostname archive path
            if c.get('symbolic_name') == 'hostname':
                self.hostname_path = os.path.join(
                    'insights_commands', mangle.mangle_command(c['command']))
            if c['command'] in rm_commands or c.get('symbolic_name') in rm_commands:
                logger.warn("WARNING: Skipping command %s", c['command'])
                BLACKLISTED_SPECS.append(c['symbolic_name'])
            elif self.mountpoint == "/" or c.get("image"):
                cmd_specs = self._parse_command_spec(c, conf['pre_commands'])
                for s in cmd_specs:
                    if s['command'] in rm_commands:
                        logger.warn("WARNING: Skipping command %s", s['command'])
                        BLACKLISTED_SPECS.append(s['symbolic_name'])
                        continue
                    cmd_spec = InsightsCommand(self.config, s, self.mountpoint)
                    self.archive.add_to_archive(cmd_spec)
                    collection_stats[s['command']] = {
                        'return_code': cmd_spec.return_code,
                        'exec_time': cmd_spec.exec_time,
                        'output_size': cmd_spec.output_size
                    }
        for f in conf['files']:
            if f['file'] in rm_files or f.get('symbolic_name') in rm_files:
                logger.warn("WARNING: Skipping file %s", f['file'])
                BLACKLISTED_SPECS.append(f['symbolic_name'])
            else:
                file_specs = self._parse_file_spec(f)
                for s in file_specs:
                    # filter files post-wildcard parsing
                    if s['file'] in rm_conf.get('files', []):
                        logger.warn("WARNING: Skipping file %s", s['file'])
                        BLACKLISTED_SPECS.append(s['symbolic_name'])
                    else:
                        file_spec = InsightsFile(s, self.mountpoint)
                        self.archive.add_to_archive(file_spec)
                        collection_stats[s['file']] = {
                            'exec_time': file_spec.exec_time,
                            'output_size': file_spec.output_size
                        }
        if 'globs' in conf:
            for g in conf['globs']:
                if g.get('symbolic_name') in rm_files:
                    # ignore glob via symbolic name
                    logger.warn("WARNING: Skipping file %s", g['glob'])
                    BLACKLISTED_SPECS.append(g['symbolic_name'])
                else:
                    glob_specs = self._parse_glob_spec(g)
                    for g in glob_specs:
                        if g['file'] in rm_files:
                            logger.warn("WARNING: Skipping file %s", g['file'])
                            BLACKLISTED_SPECS.append(g['symbolic_name'])
                        else:
                            glob_spec = InsightsFile(g, self.mountpoint)
                            self.archive.add_to_archive(glob_spec)
                            collection_stats[g['file']] = {
                                'exec_time': glob_spec.exec_time,
                                'output_size': glob_spec.output_size
                            }
        logger.debug('Spec collection finished.')

        self.redact(rm_conf)

        # collect metadata
        logger.debug('Collecting metadata...')
        self._write_branch_info(branch_info)
        self._write_display_name()
        self._write_ansible_host()
        self._write_version_info()
        self._write_tags()
        self._write_blacklist_report(blacklist_report)
        self._write_blacklisted_specs()
        self._write_egg_release()
        self._write_collection_stats(collection_stats)
        logger.debug('Metadata collection finished.')

    def redact(self, rm_conf):
        '''
        Perform data redaction (password sed command and patterns),
        write data to the archive in place
        '''
        logger.debug('Running content redaction...')

        if not re.match(r'/var/tmp/.+/insights-.+', self.archive.archive_dir):
            # sanity check to make sure we're only modifying
            #   our own stuff in temp
            # we should never get here but just in case
            raise RuntimeError('ERROR: invalid Insights archive temp path')

        if rm_conf is None:
            rm_conf = {}
        exclude = None
        regex = False
        if rm_conf:
            try:
                exclude = rm_conf['patterns']
                if isinstance(exclude, dict) and exclude['regex']:
                    # if "patterns" is a dict containing a non-empty "regex" list
                    logger.debug('Using regular expression matching for patterns.')
                    exclude = exclude['regex']
                    regex = True
                logger.warn("WARNING: Skipping patterns defined in blacklist configuration")
            except LookupError:
                # either "patterns" was undefined in rm conf, or
                #   "regex" was undefined in "patterns"
                exclude = None
        if not exclude:
            logger.debug('Patterns section of blacklist configuration is empty.')

        # TODO: consider implementing redact() in CoreCollector class rather than
        #   special handling here
        if self.config.core_collect:
            # redact only from the 'data' directory
            searchpath = os.path.join(self.archive.archive_dir, 'data')
            if not (os.path.isdir(searchpath) and
                    re.match(r'/var/tmp/.+/insights-.+/data', searchpath)):
                # abort if the dir does not exist and isn't the correct format
                # we should never get here but just in case
                raise RuntimeError('ERROR: invalid Insights archive temp path')
        else:
            searchpath = self.archive.archive_dir

        for dirpath, dirnames, filenames in os.walk(searchpath):
            for f in filenames:
                fullpath = os.path.join(dirpath, f)
                if (fullpath.endswith(
                        (
                            'etc/insights-client/machine-id',
                            'etc/machine-id',
                            'insights_commands/subscription-manager_identity',
                            'insights_commands/ls_-lanRL_.etc.systemd_.run.systemd_.usr.lib.systemd_.usr.local.lib.systemd_.usr.local.share.systemd_.usr.share.systemd',  # issue #3858
                            'var/opt/mssql/log/assessments/assessment-latest',  # issue #3885
                        )
                )):
                    # do not redact the ID files
                    continue
                redacted_contents = _process_content_redaction(fullpath, exclude, regex)
                with open(fullpath, 'wb') as dst:
                    dst.write(redacted_contents)

    def done(self, conf, rm_conf):
        """
        Do finalization stuff

        Returns:
            default:
                path to generated tarfile
            conf.obfuscate==True:
                path to generated tarfile, scrubbed by soscleaner
            conf.output_dir:
                path to a generated directory
            conf.obfuscate==True && conf.output_dir:
                path to generated directory, scubbed by soscleaner
        Ideally, we may want to have separate functions for directories
            and archive files.
        """
        if self.config.obfuscate:
            if rm_conf and rm_conf.get('keywords'):
                logger.warn("WARNING: Skipping keywords defined in blacklist configuration")
            cleaner = SOSCleaner(quiet=True)
            clean_opts = CleanOptions(
                self.config, self.archive.tmp_dir, rm_conf, self.hostname_path)
            cleaner.clean_report(clean_opts, self.archive.archive_dir)
            if clean_opts.keyword_file is not None:
                os.remove(clean_opts.keyword_file.name)

            # generate RHSM facts at this point
            self._write_rhsm_facts(cleaner.hashed_fqdn, cleaner.ip_report)

            if self.config.output_dir:
                # return the entire soscleaner dir
                #   see additions to soscleaner.SOSCleaner.clean_report
                #   for details
                return cleaner.dir_path
            else:
                # return the generated soscleaner archive
                self.archive.tar_file = cleaner.archive_path
                return cleaner.archive_path

        if self.config.output_dir:
            return self.archive.archive_dir
        else:
            return self.archive.create_tar_file()


class CleanOptions(object):
    """
    Options for soscleaner
    """
    def __init__(self, config, tmp_dir, rm_conf, hostname_path):
        self.report_dir = tmp_dir
        self.domains = []
        self.files = []
        self.quiet = True
        self.keyword_file = None
        self.keywords = None
        self.no_tar_file = config.output_dir
        self.core_collect = config.core_collect

        if rm_conf:
            try:
                keywords = rm_conf['keywords']
                self.keyword_file = NamedTemporaryFile(delete=False)
                self.keyword_file.write("\n".join(keywords).encode('utf-8'))
                self.keyword_file.flush()
                self.keyword_file.close()
                self.keywords = [self.keyword_file.name]
                logger.debug("Attmpting keyword obfuscation")
            except LookupError:
                pass

        if config.obfuscate_hostname:
            # default to its original location
            self.hostname_path = hostname_path or 'insights_commands/hostname'
        else:
            self.hostname_path = None



from insights.combiners.ipa import IPA
from insights.parsers.installed_rpms import InstalledRpms
from insights.parsers.ipa_conf import IPAConfig
from insights.parsers.redhat_release import RedhatRelease
from insights.parsers.sssd_conf import SSSD_Config

from insights.tests import context_wrap

REDHAT_RELEASE_RHEL = """
Red Hat Enterprise Linux release 9.2 (Plow)
"""

IPA_RPMS_CLIENT = """
ipa-client-4.10.1-6.el9.x86_64
"""

IPA_RPMS_SERVER = """
ipa-client-4.10.1-6.el9.x86_64
ipa-server-4.10.1-6.el9.x86_64
"""

SSSD_CONF_SERVER = """
[domain/ipa.test]
id_provider = ipa
ipa_server_mode = True
ipa_server = server.ipa.test
ipa_domain = ipa.test
ipa_hostname = server.ipa.test
auth_provider = ipa
chpass_provider = ipa
access_provider = ipa
cache_credentials = True
ldap_tls_cacert = /etc/ipa/ca.crt
krb5_store_password_if_offline = True
sudo_provider = ipa
autofs_provider = ipa
subdomains_provider = ipa
session_provider = ipa
hostid_provider = ipa

[sssd]
domains = ipa.test
"""

IPA_DEFAULT_CONF_SERVER = """
[global]
host = server.ipa.test
basedn = dc=ipahcc,dc=test
realm = ipa.test
domain = ipa.test
xmlrpc_uri = https://server.ipa.test/ipa/xml
ldap_uri = ldapi://%2Frun%2Fslapd-IPAHCC-TEST.socket
mode = production
enable_ra = True
ra_plugin = dogtag
dogtag_version = 10
"""

SSSD_CONF_CLIENT = """
[domain/ipa.test]
id_provider = ipa
ipa_server = _srv_, server.ipa.test
ipa_domain = ipa.test
ipa_hostname = client91.ipa.test
auth_provider = ipa
chpass_provider = ipa
access_provider = ipa
cache_credentials = True
ldap_tls_cacert = /etc/ipa/ca.crt
krb5_store_password_if_offline = True

[sssd]
domains = ipa.test
"""

IPA_DEFAULT_CONF_CLIENT = """
[global]
basedn = dc=ipahcc,dc=test
realm = ipa.test
domain = ipa.test
server = server.ipa.test
host = client91.ipa.test
xmlrpc_uri = https://server.ipa.test/ipa/xml
enable_ra = True
"""


def test_ipa():
    redhat_release = RedhatRelease(context_wrap(REDHAT_RELEASE_RHEL))

    rpms_client = InstalledRpms(context_wrap(IPA_RPMS_CLIENT))
    sssd_conf_client = SSSD_Config(context_wrap(SSSD_CONF_CLIENT))
    ipa_conf_client = IPAConfig(context_wrap(IPA_DEFAULT_CONF_CLIENT))

    rpms_server = InstalledRpms(context_wrap(IPA_RPMS_SERVER))
    sssd_conf_server = SSSD_Config(context_wrap(SSSD_CONF_SERVER))
    ipa_conf_server = IPAConfig(context_wrap(IPA_DEFAULT_CONF_SERVER))

    ipa_client = IPA(
        ipa_conf_client, sssd_conf_client, rpms_client, redhat_release
    )
    assert ipa_client.sssd_conf.domains == [ipa_client.ipa_conf.domain]
    assert ipa_client._is_client is None
    assert ipa_client._is_server is None
    assert ipa_client.is_client
    assert ipa_client._is_client
    assert not ipa_client.is_server
    assert not ipa_client._is_server

    ipa_server = IPA(
        ipa_conf_server, sssd_conf_server, rpms_server, redhat_release
    )
    assert ipa_server.sssd_conf.domains == [ipa_server.ipa_conf.domain]
    assert ipa_server.is_client
    assert ipa_server.is_server

    ipa_mixed = IPA(
        ipa_conf_client, sssd_conf_server, rpms_client, redhat_release
    )
    assert ipa_mixed.sssd_conf.domains == [ipa_mixed.ipa_conf.domain]
    assert ipa_mixed.is_client
    assert not ipa_mixed.is_server



# -*- coding: utf-8 -*-
"""
DseLdifSimple - file ``/etc/dirsrv/*/dse.ldif``
===============================================
"""


from insights import Parser, parser
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.dse_ldif)
class DseLdifSimple(Parser, dict):
    """
    .. warning::
        This class is deprecated and will be removed from 3.5.0.
        Please use the :class:`insights.parsers.dse_ldif.DseLDIF` instead.

    Parse the attributes out of the filtered lines of the dse.ldif file into
    dict[attribute name] = attribute value.

    Please note the difference between the LDIF format and this parser.

    The file dse.ldif is in the LDIF format (see ``man 5 ldif``). LDIF contains
    multi-row records where each record is identified by a ``dn:`` line ("dn"
    as in "distinguished name") and the record's other lines are attributes.
    Attributes may also have base64-encoded values, multiline values, and
    file-stored values.

    The parser processes lines independently without tracking what the line is
    and which record it belongs to. Only plaintext single-line values are
    supported.

    This allows for filterable, efficient, and privacy-preserving processing of
    attributes whose names are valid only in a single distinguished name, and
    whose values are single-line plaintext only.

    Sample input data::

        dn: cn=config
        nsslapd-securePort: 636
        nsslapd-security: on

        dn: cn=encryption,cn=config
        sslVersionMin: SSLv3
        sslVersionMax: TLS1.1
        nsSSL3: on

    Examples:
        >>> type(dse_ldif_simple)
        <class 'insights.parsers.dse_ldif_simple.DseLdifSimple'>
        >>> dse_ldif_simple["nsslapd-security"]
        'on'
        >>> "sslVersionMin" in dse_ldif_simple
        True
        >>> dict(dse_ldif_simple)["nsSSL3"]
        'on'

    """

    def __init__(self, *args, **kwargs):
        deprecated(DseLdifSimple, "Please use the :class:`insights.parsers.dse_ldif.DseLDIF` instead.", "3.5.0")
        super(DseLdifSimple, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        data = {}
        for line in content:
            if line.startswith("#"):
                # lines beginning with # are ignored
                continue
            if ":" not in line:
                # only attribute: value lines supported
                continue
            if line.startswith(" "):
                # multi-line values not supported
                continue
            attr_name, attr_value = line.split(":", 1)
            if attr_value.startswith(":"):
                # base64-encrypted values not supported
                continue
            if attr_value.startswith("<"):
                # file-backed values not supported
                continue

            # Whitespace at either side of the value has no effect.
            attr_value = attr_value.strip()

            # If the same attribute is declared multiple times, the first
            # instance takes into effect, the rest is ignored by the 386
            # Directory Server.
            if attr_name not in data:
                data[attr_name] = attr_value
        self.update(data)



from .. import XMLParser, parser
from insights.specs import Specs


@parser(Specs.systemid)
class SystemID(XMLParser):
    '''
    ---------------
    Return a SystemId object which contains a dict below:
        {
        "username": "testuser",
        "operating_system": "redhat-release-workstation",
        "description": "Initial Registration Parameters: OS: redhat-release-workstation Release: 6Workstation CPU Arch: x86_64",
        "checksum": "b493da72be7cfb7e54c1d58c6aa140c9",
        "profile_name": "example_profile",
        "system_id": "ID-example",
        "architecture": "x86_64",
        "os_release": "6Workstation",
        "type": "REAL"
        }
    Ignore the member with "fields" text. Because it's information is nonsense
    ---Sample---
    <?xml version="1.0"?>
    <params>
    <param>
    <value><struct>
    <member>
    <name>username</name>
    <value><string>testuser</string></value>
    </member>
    <member>
    <name>operating_system</name>
    <value><string>redhat-release-workstation</string></value>
    </member>
    <member>
    <name>description</name>
    <value><string>Initial Registration Parameters:
    OS: redhat-release-workstation
    Release: 6Workstation
    CPU Arch: x86_64</string></value>
    </member>
    <member>
    <name>checksum</name>
    <value><string>b493da72be7cfb7e54c1d58c6aa140c9</string></value>
    </member>
    <member>
    <name>profile_name</name>
    <value><string>example_profile</string></value>
    </member>
    <member>
    <name>system_id</name>
    <value><string>ID-example</string></value>
    </member>
    <member>
    <name>architecture</name>
    <value><string>x86_64</string></value>
    </member>
    <member>
    <name>os_release</name>
    <value><string>6Workstation</string></value>
    </member>
    <member>
    <name>fields</name>
    <value><array><data>
    <value><string>system_id</string></value>
    <value><string>os_release</string></value>
    <value><string>operating_system</string></value>
    <value><string>architecture</string></value>
    <value><string>username</string></value>
    <value><string>type</string></value>
    </data></array></value>
    </member>
    <member>
    <name>type</name>
    <value><string>REAL</string></value>
    </member>
    </struct></value>
    </param>
    </params>
    '''

    def parse_dom(self):
        systemid_info = {}

        for member in self.dom.findall(".//member"):
            # ignore "fields" infos
            if member[0].text != 'fields':
                key = member[0].text
                value = member[1][0].text
                systemid_info[key] = value
        return systemid_info



import pytest

from insights.core.exceptions import ParseException, SkipComponent
from insights.parsers.php_ini import PHPConf
from insights.tests import context_wrap

# Latest production php.ini from php git repository. Comments stripped out.
# https://git.php.net/?p=php-src.git;a=blob;f=php.ini-production;hb=HEAD
INI_DEFAULT = """
[PHP]
engine = On
short_open_tag = Off
precision = 14
output_buffering = 4096
zlib.output_compression = Off
implicit_flush = Off
unserialize_callback_func =
serialize_precision = -1
disable_functions =
; This is a comment.
disable_classes =
zend.enable_gc = On
zend.exception_ignore_args = On
zend.exception_string_param_max_len = 0
expose_php = On
max_execution_time = 30
max_input_time = 60
memory_limit = 128M

error_reporting = E_ALL & ~E_DEPRECATED & ~E_STRICT

display_errors = Off
display_startup_errors = Off
log_errors = On
log_errors_max_len = 1024
ignore_repeated_errors = Off
ignore_repeated_source = Off
report_memleaks = On
variables_order = "GPCS"
request_order = "GP"
register_argc_argv = Off
auto_globals_jit = On

post_max_size = 8M

auto_prepend_file =
auto_append_file =
default_mimetype = "text/html"
default_charset = "UTF-8"
doc_root =
user_dir =
enable_dl = Off
file_uploads = On
upload_max_filesize = 2M
max_file_uploads = 20
allow_url_fopen = On
allow_url_include = Off
default_socket_timeout = 60

[CLI Server]
cli_server.color = On

[Date]

[filter]

[iconv]

[imap]

[intl]

[sqlite3]

[Pcre]

[Pdo]

[Pdo_mysql]
pdo_mysql.default_socket=

[Phar]

[mail function]
SMTP = localhost
smtp_port = 25
mail.add_x_header = Off

[ODBC]
odbc.allow_persistent = On
odbc.check_persistent = On
odbc.max_persistent = -1
odbc.max_links = -1
odbc.defaultlrl = 4096
odbc.defaultbinmode = 1

[MySQLi]
mysqli.max_persistent = -1
mysqli.allow_persistent = On
mysqli.max_links = -1
mysqli.default_port = 3306
mysqli.default_socket =
mysqli.default_host =
mysqli.default_user =
mysqli.default_pw =
mysqli.reconnect = Off

[mysqlnd]
mysqlnd.collect_statistics = On
mysqlnd.collect_memory_statistics = Off

[OCI8]

[PostgreSQL]
pgsql.allow_persistent = On
pgsql.auto_reset_persistent = Off
pgsql.max_persistent = -1
pgsql.max_links = -1
pgsql.ignore_notice = 0
pgsql.log_notice = 0

[bcmath]
bcmath.scale = 0

[browscap]

[Session]
session.save_handler = files
session.use_strict_mode = 0
session.use_cookies = 1
session.use_only_cookies = 1
session.name = PHPSESSID
session.auto_start = 0
session.cookie_lifetime = 0
session.cookie_path = /
session.cookie_domain =
session.cookie_httponly =
session.cookie_samesite =
session.serialize_handler = php
session.gc_probability = 1
session.gc_divisor = 1000
session.gc_maxlifetime = 1440
session.referer_check =
session.cache_limiter = nocache
session.cache_expire = 180
session.use_trans_sid = 0
session.sid_length = 26
session.trans_sid_tags = "a=href,area=href,frame=src,form="
session.sid_bits_per_character = 5

[Assertion]
zend.assertions = -1

[COM]

[mbstring]

[gd]

[exif]

[Tidy]
tidy.clean_output = Off

[soap]
soap.wsdl_cache_enabled=1
soap.wsdl_cache_dir="/tmp"
soap.wsdl_cache_ttl=86400
soap.wsdl_cache_limit = 5

[sysvshm]

[ldap]
ldap.max_links = -1

[dba]

[opcache]

[curl]

[openssl]

[ffi]

""".strip()

INI_EMPTY = ""
INI_INVALID = "bla bla foo ha [] ^&*@#$%"


def test_php_conf_default():
    php_c = PHPConf(context_wrap(INI_DEFAULT))
    assert php_c['PHP']['default_mimetype'].value == 'text/html'
    assert php_c.data['PHP']['default_mimetype'] == 'text/html'
    assert php_c.data['Session']['session.cache_limiter'] == 'nocache'
    assert php_c['PHP']['engine'].value is True
    assert php_c['PHP']['precision'].value == 14
    assert php_c['PHP']['disable_classes'].value == ''
    assert php_c['PHP']['memory_limit'].value == 128 * 2**20  # Conversion of 128M to bytes.
    assert php_c['PHP']['post_max_size'].value == 8 * 2**20  # Conversion of 8M to bytes.


def test_php_conf_empty():
    with pytest.raises(SkipComponent):
        PHPConf(context_wrap(INI_EMPTY))


def test_php_conf_invalid():
    with pytest.raises(ParseException):
        PHPConf(context_wrap(INI_INVALID))



"""
CloudInitCustomeNetwork - file ``/etc/cloud/cloud.cfg.d/99-custom-networking.cfg``
==================================================================================

This module provides parsing for cloudinit custom networking configuration file.
``CloudInitCustomNetworking`` is a parser for ``/etc/cloud/cloud.cfg.d/99-custom-networking.cfg`` files.

Typical output is::

    network:
      version: 1
      config:
      - type: physical
        name: eth0
        subnets:
          - type: dhcp
          - type: dhcp6

Examples:
    >>> cloud_init_custom_network_config.data['network']['config'][0]['name']
    'eth0'
    >>> cloud_init_custom_network_config.data['network']['config'][0]['subnets'][0]['type'] == 'dhcp'
    True
    >>> cloud_init_custom_network_config.data['network']['config'][0]['subnets'][1]['type'] == 'dhcp6'
    True
"""


from insights import YAMLParser, parser
from insights.specs import Specs


@parser(Specs.cloud_init_custom_network)
class CloudInitCustomNetworking(YAMLParser):
    """ Class for parsing the content of ``/etc/cloud/cloud.cfg.d/99-custom-networking.cfg``."""
    pass



"""
LuksDump - command ``cryptsetup luksDump``
==========================================
This class provides parsing for the output of cryptsetup luksDump
<device_name>. Outputs from LUKS1 and LUKS2 are supported.
"""
import string

from insights.core import Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsr import (AnyChar, Char, EOF, HangingString, Lift,
                            Literal, Many, Opt, String, WithIndent, WS)
from insights.specs import Specs


class DocParser(object):
    def __init__(self):
        value_chars = set(string.printable) - set("\n\r")

        FirstLine = Literal("LUKS header information", value="header") << AnyChar.until(Char("\n")) + Opt(Many(Char("\n")))
        FirstIndent = Literal("  ")
        # we need to replace the \t by 8 spaces in the input,
        # otherwise WithIndent does not work properly
        # SecondIndent = Literal("\t")
        SecondIndent = Literal(" " * 8)

        Key = String(value_chars - set(":")) << Char(":") % "Key"
        Value = WS >> HangingString(value_chars) % "Value"

        MultilineContinuation = Many((Char("\n") + Literal(" " * 15) + SecondIndent) >> String(value_chars))
        Value1 = WS >> String(value_chars) + Opt(MultilineContinuation).map(lambda x: "".join(x)) << Char("\n")
        Value1 = Value1.map(lambda x: ("".join(x)).strip())

        ZeroLevelKVPair = Key + Value1
        FirstLevelKVPair = FirstIndent >> Key + Value1
        SecondLevelKVPair = SecondIndent >> WithIndent(Key + Value) << Opt(Many(Char("\n")))

        Luks2SectionName = Key << Char("\n")
        Luks2SectionEntry = (FirstLevelKVPair + Many(SecondLevelKVPair).map(dict)).map(self.convert_type)
        Luks2Section = Luks2SectionName + Many(Luks2SectionEntry).map(dict) << Opt(Many(Char("\n")))
        Luks2Body = Many(Luks2Section, lower=1)

        Luks1Section = ZeroLevelKVPair + Many(SecondLevelKVPair).map(dict) << Opt(Many(Char("\n")))
        Luks1Body = Many(Luks1Section.map(self.convert_status), lower=1)

        KVBlock = Many(Key + Value1).map(dict)
        LuksHeader = (FirstLine + KVBlock) << Opt(Many(Char("\n")))

        # Luks2Body has to go first, because Luks1Body consumes also valid Luks2 bodies
        self.Top = Lift(self.join_header_and_body) * LuksHeader * (Luks2Body | Luks1Body) << EOF

    def join_header_and_body(self, header, body):
        return dict([header] + body)

    def convert_type(self, section):
        section[1]["type"] = section[0][1]
        return [section[0][0], section[1]]

    def convert_status(self, section):
        section[2]["status"] = section[1]
        return [section[0], section[2]]

    def __call__(self, content):
        try:
            return self.Top(content)
        except Exception:
            raise ParseException("There was an exception when parsing one of the outputs of cryptsetup luksDump commands.")


@parser(Specs.cryptsetup_luksDump)
class LuksDump(Parser):
    """
    Sample input data is in the format::

        LUKS header information
        Version:       	2
        Epoch:         	6
        Metadata area: 	16384 [bytes]
        Keyslots area: 	16744448 [bytes]
        UUID:          	cfbcc942-e06b-4c4a-952f-e9c9b2011c27
        Label:         	(no label)
        Subsystem:     	(no subsystem)
        Flags:       	(no flags)

        Data segments:
          0: crypt
                offset: 16777216 [bytes]
                length: (whole device)
                cipher: aes-xts-plain64
                sector: 4096 [bytes]

        Keyslots:
          0: luks2
                Key:        512 bits
                Priority:   normal
                Cipher:     aes-xts-plain64
                Cipher key: 512 bits
                PBKDF:      argon2id
                Time cost:  7
                Memory:     1048576
                Threads:    4
                Salt:       3d c4 1b 52 fe 1c 90 d8 2a 35 b2 62 34 e9 0a 59 
                            e9 0e 48 57 b2 dd 45 
                AF stripes: 4000
                AF hash:    sha256
                Area offset:32768 [bytes]
                Area length:258048 [bytes]
                Digest ID:  0
        Tokens:
          0: systemd-tpm2
                Keyslot:    2
        Digests:
          0: pbkdf2
                Hash:       sha256
                Iterations: 129774
                Salt:       e6 31 d5 74 e0 65 83 82 35 03 29 56 0e 80 36 5c 
                            4d cd 4d f9 de 69 39 97 
                Digest:     21 aa b3 dc 9d 46 9b 0f 3a 0f 57 13 80 c6 0b bf 
                            67 66 9e 73 ed 7d 09 2c 

    Examples:
        >>> type(parsed_result)
        <class 'insights.parsers.cryptsetup_luksDump.LuksDump'>

        >>> from pprint import pprint
        >>> pprint(parsed_result.dump["header"])
        {'Epoch': '6',
         'Flags': '(no flags)',
         'Keyslots area': '16744448 [bytes]',
         'Label': '(no label)',
         'Metadata area': '16384 [bytes]',
         'Subsystem': '(no subsystem)',
         'UUID': 'cfbcc942-e06b-4c4a-952f-e9c9b2011c27',
         'Version': '2'}

        >>> pprint(parsed_result.dump["Keyslots"]["0"])
        {'AF hash': 'sha256',
         'AF stripes': '4000',
         'Area length': '258048 [bytes]',
         'Area offset': '32768 [bytes]',
         'Cipher': 'aes-xts-plain64',
         'Cipher key': '512 bits',
         'Digest ID': '0',
         'Key': '512 bits',
         'Memory': '1048576',
         'PBKDF': 'argon2id',
         'Priority': 'normal',
         'Salt': '3d c4 1b 52 fe 1c 90 d8 2a 35 b2 62 34 e9 0a 59 e9 0e 48 57 b2 dd 45',
         'Threads': '4',
         'Time cost': '7',
         'type': 'luks2'}

        >>> parsed_result.dump["Tokens"]["0"]["type"]
        'systemd-tpm2'


    Attributes:
        dump(dict of dicts): A top level dict containing the dictionaries
            representing the header, data segments, keyslots, digests
            and tokens.
    """  # noqa

    def __init__(self, context):
        self.parse_dump = DocParser()
        super(LuksDump, self).__init__(context)

    def parse_content(self, content):
        if len(content) == 0 or (len(content) == 1 and "not a valid LUKS" in content[0]):
            raise SkipComponent
        self.dump = self.parse_dump("\n".join(content).replace("\t", " " * 8) + "\n")



"""
rhsm.conf - File /etc/rhsm/rhsm.conf
====================================
"""
from insights.core import IniConfigFile
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.specs import Specs

add_filter(Specs.rhsm_conf, '[')


@parser(Specs.rhsm_conf)
class RHSMConf(IniConfigFile):
    """
    Parses content of "/etc/rhsm/rhsm.conf".

    Typical content of "/etc/rhsm/rhsm.conf" is::

        # Unified Entitlement Platform Configuration
        [server]
        # Server hostname:
        hostname = subscription.rhn.redhat.com

        # Server prefix:
        prefix = /subscription

        # Server port:
        port = 443

        # Set to 1 to disable certificate validation:
        insecure = 0

        # Set the depth of certs which should be checked
        # when validating a certificate
        ssl_verify_depth = 3

        # an http proxy server to use
        proxy_hostname =

        # port for http proxy server
        proxy_port =

        [rhsm]
        # Content base URL:
        baseurl= https://cdn.redhat.com

        # Server CA certificate location:
        ca_cert_dir = /etc/rhsm/ca/

        # Default CA cert to use when generating yum repo configs:
        repo_ca_cert = %(ca_cert_dir)sredhat-uep.pem

        # Where the certificates should be stored
        productCertDir = /etc/pki/product
        entitlementCertDir = /etc/pki/entitlement
        consumerCertDir = /etc/pki/consumer

        # Manage generation of yum repositories for subscribed content:
        manage_repos = 1

        # Refresh repo files with server overrides on every yum command
        full_refresh_on_yum = 0

        # If set to zero, the client will not report the package profile to
        # the subscription management service.
        report_package_profile = 1

        # The directory to search for subscription manager plugins
        pluginDir = /usr/share/rhsm-plugins

        # The directory to search for plugin configuration files
        pluginConfDir = /etc/rhsm/pluginconf.d

        [rhsmcertd]
        # Interval to run cert check (in minutes):
        certCheckInterval = 240
        # Interval to run auto-attach (in minutes):
        autoAttachInterval = 1440

    Examples:
        >>> type(conf)
        <class 'insights.parsers.rhsm_conf.RHSMConf'>
        >>> conf.sections()
        ['server', 'rhsm', 'rhsmcertd']
        >>> conf.has_option('rhsm', 'ca_cert_dir')
        True
        >>> conf.get("rhsm", "baseurl")
        'https://cdn.redhat.com'
        >>> conf.get("rhsm", "pluginDir")
        '/usr/share/rhsm-plugins'
        >>> conf.getboolean("rhsm", "manage_repos")
        True
    """
    pass



from insights.parsers.openvswitch_other_config import OpenvSwitchOtherConfig
from insights.tests import context_wrap

other_config_1 = """
{dpdk-init="true", dpdk-lcore-mask="30000003000", dpdk-socket-mem="4096,4096", pmd-cpu-mask="30000003000"}
""".strip()

other_config_2 = """
{dpdk-init="true", dpdk-lcore-mask="c00000c00", dpdk-socket-mem="4096,4096", pmd-cpu-mask="c00000c00"}
""".strip()

other_config_3 = """
{}
""".strip()

other_config_4 = """
{pmd-cpu-mask="3000000000030030030"}
""".strip()

other_config_5 = """
ovs-vsctl: unix:/var/run/openvswitch/db.sock: database connection failed (Permission denied)
""".strip()


def test_openvswitch_():
    result = OpenvSwitchOtherConfig(context_wrap(other_config_1))
    assert "pmd-cpu-mask" in result
    assert result.get("dpdk-init") == "true"
    assert result.get("dpdk-socket-mem") == "4096,4096"
    assert result.get("pmd-cpu-mask") == "30000003000"
    assert result.get("dpdk-lcore-mask") == "30000003000"
    assert result["dpdk-lcore-mask"] == "30000003000"

    result = OpenvSwitchOtherConfig(context_wrap(other_config_2))
    assert result.get("dpdk-init") == "true"
    assert result.get("dpdk-socket-mem") == "4096,4096"
    assert result.get("pmd-cpu-mask") == "c00000c00"

    result = OpenvSwitchOtherConfig(context_wrap(other_config_3))
    assert result.get("dpdk-init") is None

    result = OpenvSwitchOtherConfig(context_wrap(other_config_4))
    assert result.get("dpdk-init") is None
    assert result.get("pmd-cpu-mask") == "3000000000030030030"

    result = OpenvSwitchOtherConfig(context_wrap(other_config_5))
    assert result.get("dpdk-init") is None



#!/usr/bin/env python

from __future__ import print_function

import re
import socket
import uuid

from insights import rule, run, make_metadata
from insights.combiners.cloud_instance import CloudInstance
from insights.combiners.cloud_provider import CloudProvider
from insights.core import Parser
from insights.core.dr import set_enabled, load_components
from insights.core.plugins import parser
from insights.parsers.aws_instance_id import AWSInstanceIdDoc
from insights.parsers.azure_instance import AzureInstanceID, AzureInstanceType
from insights.parsers.dmidecode import DMIDecode
from insights.parsers.gcp_instance_type import GCPInstanceType
from insights.parsers.installed_rpms import InstalledRpms
from insights.parsers.rhsm_conf import RHSMConf
from insights.parsers.subscription_manager import SubscriptionManagerID, SubscriptionManagerFacts
from insights.parsers.yum import YumRepoList
from insights.specs import Specs


def valid_uuid_or_None(s):
    """
    Returns  if `s` is a valid UUID string.
    """
    if not s:
        return None
    try:
        return str(uuid.UUID(s))
    except Exception:
        return None


def valid_ipv4_address_or_None(addr):
    """ str: Returns the input value if it is a valid IPV4 address """
    try:
        socket.inet_pton(socket.AF_INET, addr)
        return addr
    except socket.error:
        return None


def valid_mac_addresses(mac_address_datasources):
    """ list: Return a list of valid mac addresses from a list of datasources """
    valid_addrs = []
    for ds in mac_address_datasources:
        try:
            addr = ds.content[0].strip()
        except Exception:
            continue
        # Only look for addresses in the form 00:00:00:00:00:00
        match = re.match("^([0-9a-f]{2}:){5}[0-9a-f]{2}$", addr)
        if match is not None:
            valid_addrs.append(addr)

    return valid_addrs


@parser(Specs.ip_addresses)
class IPs(Parser):
    """
    Reads the output of ``hostname -I`` and constructs a list of all assigned IP
    addresses. This command should only output IPV4 addresses and should not
    include localhost, but sometimes it does.  The validation function removes
    those from the list.

    Example output::
       192.168.1.71 10.88.0.1 172.17.0.1 172.18.0.1 10.10.121.131

    Resultant data structure::
        [
            "192.168.1.71",
            "10.88.0.1",
            "172.17.0.1",
            "172.18.0.1",
            "10.10.121.131",
        ]
    """
    def parse_content(self, content):
        self.data = list(filter(None, [valid_ipv4_address_or_None(addr) for addr in content[0].rstrip().split()]))


def _safe_parse(ds):
    try:
        return ds.content[0]

    except Exception:
        return None


def _filter_falsy(dict_):
    return dict((k, v) for k, v in dict_.items() if v)


@rule(
    optional=[
        Specs.machine_id,
        Specs.etc_machine_id,
        Specs.bios_uuid,
        SubscriptionManagerID,
        IPs,
        Specs.hostname,
        Specs.mac_addresses,
        CloudInstance,
    ]
)
def canonical_facts(
    insights_id, machine_id, bios_uuid, submanid, ips, fqdn, mac_addresses,
    cloud_instance,
):

    facts = dict(
        insights_id=valid_uuid_or_None(_safe_parse(insights_id)),
        machine_id=valid_uuid_or_None(_safe_parse(machine_id)),
        bios_uuid=valid_uuid_or_None(_safe_parse(bios_uuid)),
        subscription_manager_id=valid_uuid_or_None(submanid.identity if submanid else None),
        ip_addresses=ips.data if ips else [],
        mac_addresses=valid_mac_addresses(mac_addresses) if mac_addresses else [],
        fqdn=_safe_parse(fqdn),
        provider_id=cloud_instance.id if cloud_instance else None,
        provider_type=cloud_instance.provider if cloud_instance else None,
    )

    return make_metadata(**_filter_falsy(facts))


def get_canonical_facts(path=None):
    load_components("insights.specs.default", "insights.specs.insights_archive")

    required_components = [
        AWSInstanceIdDoc,
        AzureInstanceID,
        AzureInstanceType,
        DMIDecode,
        GCPInstanceType,
        IPs,
        InstalledRpms,
        RHSMConf,
        SubscriptionManagerFacts,
        SubscriptionManagerID,
        YumRepoList,
        CloudProvider,
        CloudInstance,
        canonical_facts,
    ]
    for comp in required_components:
        set_enabled(comp, True)

    br = run(canonical_facts, root=path)
    d = br[canonical_facts]
    del d["type"]
    return d


if __name__ == "__main__":
    import json

    print(json.dumps(get_canonical_facts()))



import pytest

from insights.core.exceptions import SkipComponent
from insights.combiners.krb5 import AllKrb5Conf
from insights.combiners.identity_domain import (
    IdentityDomain,
    DomainInfo,
    DomainTypes,
    ClientSoftware,
    ServerSoftware,
    IPAMode,
)
from insights.combiners.ipa import IPA
from insights.parsers.installed_rpms import InstalledRpms
from insights.parsers.ipa_conf import IPAConfig
from insights.parsers.krb5 import Krb5Configuration
from insights.parsers.redhat_release import RedhatRelease
from insights.parsers.samba import SambaConfigs
from insights.parsers.sssd_conf import SSSD_Config
from insights.tests import context_wrap

KRB5_CONF = """
[libdefaults]
    dns_lookup_realm = false
    default_realm = IPA.TEST
    dns_canonicalize_hostname = false
    ticket_lifetime = 24h
    renew_lifetime = 7d
    rdns = false

[realms]
    IPA.TEST = {
        pkinit_anchors = FILE:/var/lib/ipa-client/pki/kdc-ca-bundle.pem
        pkinit_pool = FILE:/var/lib/ipa-client/pki/ca-bundle.pem
    }
    KERBEROS-LDAP.TEST = {
        kdc = kdc.kerberos-ldap.test
    }

    KERBEROS.TEST = {
        kdc = kdc.kerberos.test
    }

    # "realm join" for AD  does not add an entry to krb5.conf.
    # AD-WINBIND.TEST = {}
    # AD-SSSD.TEST = {}

[domain_realm]
  .ipa.test = IPA.TEST
  ipa.test = IPA.TEST
"""

SMB_CONF_WINBIND = """
# output of testparm -s
Server role: ROLE_DOMAIN_MEMBER

[global]
    realm = AD-WINBIND.TEST
    security = ADS
    template homedir = /home/%U@%D
    template shell = /bin/bash
    winbind offline logon = Yes
    winbind refresh tickets = Yes
    workgroup = AD-WINBIND
    idmap config * : range = 10000-999999
    idmap config sub1 : backend = rid
    idmap config sub1 : range = 2000000-2999999
    idmap config * : backend = tdb
"""

REDHAT_RELEASE_RHEL = """
Red Hat Enterprise Linux release 9.2 (Plow)
"""

IPA_RPMS_CLIENT = """
ipa-client-4.10.1-6.el9.x86_64
"""

IPA_DEFAULT_CONF = """
[global]
basedn = dc=ipa,dc=test
realm = IPA.TEST
domain = ipa.test
server = server.ipa.test
xmlrpc_uri = https://server.ipa.test/ipa/xml
enable_ra = True
"""

SSSD_CONF = """
[sssd]
domains = ad-sssd.test, ipa.test, kerberos-ldap, ldap
config_file_version = 2
services = nss, pam

[domain/ad-sssd.test]
default_shell = /bin/bash
krb5_store_password_if_offline = True
cache_credentials = True
krb5_realm = AD-SSSD.TEST
realmd_tags = manages-system joined-with-adcli
id_provider = ad
fallback_homedir = /home/%u@%d
ad_domain = ad-sssd.test
use_fully_qualified_names = True
ldap_id_mapping = True
access_provider = ad

[domain/ipa.test]
id_provider = ipa
ipa_server = _srv_, server.ipa.test
ipa_domain = ipa.test
ipa_hostname = client91.ipa.test
auth_provider = ipa
chpass_provider = ipa
access_provider = ipa
cache_credentials = True
ldap_tls_cacert = /etc/ipa/ca.crt
krb5_store_password_if_offline = True

[domain/kerberos-ldap]
id_provider = ldap
ldap_uri = ldap://ldap.kerberos-ldap.test
ldap_search_base = dc=kerberos-ldap,dc=test
auth_provider = krb5
krb5_server = kdc.kerberos-ldap.test
krb5_realm = KERBEROS-LDAP.TEST

[domain/ldap]
id_provider = ldap
ldap_uri = ldap://ldap.ldap.test
ldap_search_base = dc=ldap,dc=test?subtree?
cache_credentials = true
"""

IPA_DOMAIN = DomainInfo(
    "ipa.test",
    DomainTypes.IPA,
    ServerSoftware.IPA,
    ClientSoftware.SSSD,
    "ipa.test",
    "IPA.TEST",
    None,
    IPAMode.IPA_CLIENT,
)

KERBEROS_LDAP_DOMAIN = DomainInfo(
    "kerberos-ldap",
    DomainTypes.LDAP_KRB5,
    ServerSoftware.LDAP_KRB5,
    ClientSoftware.SSSD,
    None,
    "KERBEROS-LDAP.TEST",
    None,
    None,
)

LDAP_DOMAIN = DomainInfo(
    "ldap",
    DomainTypes.LDAP,
    ServerSoftware.LDAP,
    ClientSoftware.SSSD,
    None,
    None,
    None,
    None,
)

AD_SSSD_DOMAIN = DomainInfo(
    "ad-sssd.test",
    DomainTypes.AD_SSSD,
    ServerSoftware.AD,
    ClientSoftware.SSSD,
    "ad-sssd.test",
    "AD-SSSD.TEST",
    None,
    None,
)

KERBEROS_DOMAIN = DomainInfo(
    "kerberos.test",
    DomainTypes.KRB5,
    ServerSoftware.KRB5,
    ClientSoftware.KRB5,
    None,
    "KERBEROS.TEST",
    None,
    None,
)

AD_WINBIND_DOMAIN = DomainInfo(
    "ad-winbind.test",
    DomainTypes.AD_WINBIND,
    ServerSoftware.AD,
    ClientSoftware.WINBIND,
    "ad-winbind.test",
    "AD-WINBIND.TEST",
    "AD-WINBIND",
    None,
)


def test_identity_domain_sssd():
    krb5 = AllKrb5Conf(
        [Krb5Configuration(context_wrap(KRB5_CONF, path="/etc/krb5.conf"))]
    )
    sssd = SSSD_Config(context_wrap(SSSD_CONF, path="/etc/sssd/sssd.conf"))
    redhat_release = RedhatRelease(context_wrap(REDHAT_RELEASE_RHEL))
    rpms_client = InstalledRpms(context_wrap(IPA_RPMS_CLIENT))
    ipa_conf = IPAConfig(context_wrap(IPA_DEFAULT_CONF, path="/etc/ipa/default.conf"))
    ipa = IPA(ipa_conf, sssd, rpms_client, redhat_release)

    identity_domain = IdentityDomain(sssd, krb5, ipa, None)

    assert identity_domain.domains == [
        AD_SSSD_DOMAIN,
        IPA_DOMAIN,
        KERBEROS_LDAP_DOMAIN,
        LDAP_DOMAIN,
        KERBEROS_DOMAIN,
    ]

    assert identity_domain.default_realm == "IPA.TEST"
    assert identity_domain.dns_lookup_kdc is True
    assert identity_domain.dns_lookup_realm is False


def test_identity_domain_samba():
    smbwb = SambaConfigs(context_wrap(SMB_CONF_WINBIND, path="/etc/samba/smb.conf"))
    identity_domain = IdentityDomain(None, None, None, smbwb)
    assert identity_domain.domains == [AD_WINBIND_DOMAIN]

    assert identity_domain.default_realm is None
    assert identity_domain.dns_lookup_kdc is True
    assert identity_domain.dns_lookup_realm is True


def test_identity_domain_empty():
    with pytest.raises(SkipComponent):
        IdentityDomain(None, None, None, None)



"""
FindmntPropagation - command ``findmnt -lo+PROPAGATION``
========================================================

This module provides status of propagation flag of filesystems using the output
of command ``findmnt -lo+PROPAGATION``.
"""
from insights.core import CommandParser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.parsers import keyword_search, parse_fixed_table
from insights.specs import Specs


@parser(Specs.findmnt_lo_propagation)
class FindmntPropagation(CommandParser):
    """Parse output of ``findmnt -lo+PROPAGATION``.

    Typical output of ``findmnt -lo+PROPAGATION`` command is::

        TARGET                                                          SOURCE                                FSTYPE          OPTIONS                                                                       PROPAGATION
        /sys                                                            sysfs                                 sysfs           rw,nosuid,nodev,noexec,relatime,seclabel                                      shared
        /proc                                                           proc                                  proc            rw,nosuid,nodev,noexec,relatime                                               shared
        /dev                                                            devtmpfs                              devtmpfs        rw,nosuid,seclabel,size=8035516k,nr_inodes=2008879,mode=755                   shared
        /sys/kernel/security                                            securityfs                            securityfs      rw,nosuid,nodev,noexec,relatime                                               shared
        /dev/shm                                                        tmpfs                                 tmpfs           rw,nosuid,nodev,seclabel                                                      shared
        /run/netns                                                      tmpfs[/netns]                         tmpfs           rw,nosuid,nodev,seclabel,mode=755                                             shared
        /netns                                                          tmpfs[/netns]                         tmpfs           rw,nosuid,nodev,seclabel,mode=755                                             shared
        /run/netns/qdhcp-08f32dab-927e-4a61-933d-57d425827b57           proc                                  proc            rw,nosuid,nodev,noexec,relatime                                               shared
        /run/netns/qdhcp-fd138c0a-5ec7-44f8-88df-0501c4c7a968           proc                                  proc            rw,nosuid,nodev,noexec,relatime                                               shared


    Examples:

        >>> output.search_target('shm') == [{'target': '/dev/shm', 'source': 'tmpfs', 'fstype': 'tmpfs', 'options': 'rw,nosuid,nodev,seclabel', 'propagation': 'shared'}]
        True
        >>> len(output.target_startswith('/run/netns')) == 3
        True
        >>> output.target_startswith('/run/netns')[0].get('propagation', None) == 'shared'
        True

    Attributes:
        cols (list): List of key value pair derived from the command.

        keywords (list): keywords(or TARGETs) present in the command
    """
    def parse_content(self, content):
        self.cols = []
        self.keywords = []
        if not content:
            raise SkipComponent("No data.")

        self.cols = parse_fixed_table(content,
                                      header_substitute=[
                                          ('TARGET', 'target'),
                                          ('SOURCE', 'source'),
                                          ('FSTYPE', 'fstype'),
                                          ('OPTIONS', 'options'),
                                          ('PROPAGATION', 'propagation')])
        self.keywords = [col['target'] for col in self.cols]

    def search_target(self, target):
        """Similar to __contains__() but returns the list of targets.

        Example:

            >>> output.search_target('shm') == [{'target': '/dev/shm', 'source': 'tmpfs', 'fstype': 'tmpfs', 'options': 'rw,nosuid,nodev,seclabel', 'propagation': 'shared'}]
            True
        """
        return keyword_search(self.cols, target__contains=target)

    def target_startswith(self, target):
        """Return all the targets that starts with 'target'. Useful to find the mountpoints.

        Example:

            >>> len(output.target_startswith('/run/netns')) == 3
            True
        """
        return keyword_search(self.cols, target__startswith=target)
